"""
A single home standing in the middle of nowhere with
two agents living inside it as family
"""

from copy import deepcopy

from piperabm.infrastructure.samples.infrastructure_0 import model as model_infrastructure


model = deepcopy(model_infrastructure)
model.set_seed(2)
model.society.generate(
    num=2,
    gini_index=0.45,
    average_balance=1000
)
model.set_seed(None)


if __name__ == "__main__":
    print("gini index: ", model.society.gini_index)
    #model.show()


"""
Serialization for networkx objects that includes both node and edge attributes
"""

import networkx as nx


terminology = {
    'type': 't',
    'nodes': 'n',
    'edges': 'e',
    'MultiDiGraph': 'MDG',
    'MultiGraph': 'MG',
    'DiGraph': 'DG',
    'Graph': 'G'
}


def nx_serialize(G) -> dict:
    """
    Serialize networkx object
    """
    result = {}

    # Type
    if isinstance(G, nx.MultiDiGraph):
        type = terminology["MultiDiGraph"]
        multi = True
    elif isinstance(G, nx.MultiGraph):
        type = terminology["MultiGraph"]
        multi = True
    elif isinstance(G, nx.DiGraph):
        type = terminology["DiGraph"]
        multi = False
    elif isinstance(G, nx.Graph):
        type = terminology["Graph"]
        multi = False
    else:
        raise TypeError
    result[terminology['type']] = type # Type

    # Nodes
    nodes_serialized = {}
    for node in G.nodes():
        nodes_serialized[node] = G.nodes[node]
    result[terminology['nodes']] = nodes_serialized

    # Edges 
    edges_serialized = {}
    if multi is False:
        for u, v, data in G.edges(data=True):
            if u not in edges_serialized:
                edges_serialized[u] = {}
            edges_serialized[u][v] = data
    elif multi is True:
        for u, v, key, data in G.edges(keys=True, data=True):
            if u not in edges_serialized:
                edges_serialized[u] = {}
            if v not in edges_serialized[u]:
                edges_serialized[u][v] = {}
            edges_serialized[u][v][key] = data
    result[terminology['edges']] = edges_serialized # Edges

    return result


def nx_deserialize(dictionary: dict):
    """
    Deserialize networkx object
    """

    # Type
    if dictionary[terminology['type']] == terminology["MultiDiGraph"]:
        G = nx.MultiDiGraph()
        multi = True
    elif dictionary[terminology['type']] == terminology["MultiGraph"]:
        G = nx.MultiGraph()
        multi = True
    elif dictionary[terminology['type']] == terminology["DiGraph"]:
        G = nx.DiGraph()
        multi = False
    elif dictionary[terminology['type']] == terminology["Graph"]:
        G = nx.Graph()
        multi = False
    else:
        raise TypeError
    
    # Nodes
    nodes_serialized = dictionary[terminology['nodes']]
    for node in nodes_serialized:
        G.add_node(node)
        for key in nodes_serialized[node]:
            G.nodes[node][key] = nodes_serialized[node][key]

    # Edges
    edges_serialized = dictionary[terminology['edges']]
    for u in edges_serialized:
        edge_serialized_u = edges_serialized[u]
        for v in edge_serialized_u:
            if multi is False:
                edge_attributes = edge_serialized_u[v]
                G.add_edge(u, v, **edge_attributes)
            elif multi is True:
                edge_serialized_u_v = edge_serialized_u[v]
                for key in edge_serialized_u_v:
                    edge_attributes = edge_serialized_u_v[key]
                    G.add_edge(u, v, **edge_attributes)

    return G


if __name__ == "__main__":
    G = nx.Graph()
    G.add_node(1, weight=1)
    G.add_node(2, weight=2)
    G.add_edge(1, 2, weight=3)
    G_serialized = nx_serialize(G)
    print(G_serialized)
    G_new = nx_deserialize(G_serialized)
    print(nx_serialize(G_new))
    print("Test: ", nx_serialize(G_new) == G_serialized)


from copy import deepcopy

from piperabm.infrastructure.samples.infrastructure_2 import model as model_infrastructure


model = deepcopy(model_infrastructure)
model.set_seed(2)
model.society.neighbor_radius = 270
model.society.generate(
    num=10,
    gini_index=0.45,
    average_resources={
        'food': 10,
        'water': 10,
        'energy': 10,
    },
    average_balance=100
)
model.set_seed(None)


if __name__ == "__main__":
    print(model.society)
    print("society serialized: ", model.society.serialize())
    #model.show()


import numpy as np

from piperabm.tools.gini import gini
from piperabm.society.info import *


class Generate:
    """
    Generate agents
    """

    def generate(
            self,
            num: int = 1,
            gini_index: float = 0,
            average_resources: dict = {
                'food': 10,
                'water': 10,
                'energy': 10,
            },
            average_balance: float = 0
        ):
        """
        Generate agents
        """
        distribution = gini.lognorm(gini_index)
        socioeconomic_status_values = distribution.rvs(
            sample_size=num,
            percision=0.03
        )
        homes_id = self.infrastructure.homes
        for value in socioeconomic_status_values:
            socioeconomic_status = float(value)
            home_id = int(np.random.choice(homes_id))
            resources = {}
            for name in average_resources:
                resources[name] = average_resources[name] * socioeconomic_status
            balance = average_balance * socioeconomic_status 
            self.add_agent(
                home_id=home_id,
                socioeconomic_status=socioeconomic_status,
                resources=resources,
                balance=balance
            )


if __name__ == "__main__":

    from piperabm.infrastructure.samples.infrastructure_0 import model

    model.set_seed(2)
    model.society.generate(
        gini_index=0.45,
        num=2,
        average_balance=1000
    )
    model.set_seed(None)

    print("gini index: ", model.society.gini_index)
    print("society serialized:\n", model.society.serialize())
    


from piperabm.tools.coordinate import distance as ds
from piperabm.tools.coordinate import intersect
from piperabm.infrastructure.grammar.rules.rule_1 import Rule1


class Rule2:
    """
    Condition for node to edge proximity
    """

    name = "rule 2"

    def __init__(self, infrastructure, proximity_radius: float = 1):
        self.infrastructure = infrastructure
        self.proximity_radius = proximity_radius

    def check(self, edge_ids, other_edge_ids):
        result = False
        edge_pos_1 = self.infrastructure.get_pos(edge_ids[0])
        edge_pos_2 = self.infrastructure.get_pos(edge_ids[1])
        other_edge_pos_1 = self.infrastructure.get_pos(other_edge_ids[0])
        other_edge_pos_2 = self.infrastructure.get_pos(other_edge_ids[1])
        intersection = intersect.line_line(
            edge_pos_1,
            edge_pos_2,
            other_edge_pos_1,
            other_edge_pos_2
        )

        # Check if the edges are not parallel
        if intersection is not None:
            distance_1 = ds.point_to_point(intersection, edge_pos_1)
            distance_2 = ds.point_to_point(intersection, edge_pos_2)
            other_distance_1 = ds.point_to_point(intersection, other_edge_pos_1)
            other_distance_2 = ds.point_to_point(intersection, other_edge_pos_2)
            length = self.infrastructure.get_edge_attribute(ids=edge_ids, attribute='length')
            other_length = self.infrastructure.get_edge_attribute(ids=other_edge_ids, attribute='length')

            # Check if the intersection is inside the segments
            if distance_1 < length and \
                distance_2 < length and \
                other_distance_1 < other_length and \
                other_distance_2 < other_length:

                # Check if the intersection is out of the both ends of both edges
                if distance_1 > self.proximity_radius and \
                    distance_2 > self.proximity_radius and \
                    other_distance_1 > self.proximity_radius and \
                    other_distance_2 > self.proximity_radius:

                    result = True
        
        return result, intersection
    
    def apply(self, edge_ids, other_edge_ids, intersection, report=False):
        new_node_id = self.infrastructure.add_junction(pos=intersection, report=report)
        rule_1 = Rule1(self.infrastructure, proximity_radius=self.proximity_radius)
        rule_1.apply(node_id=new_node_id, edge_ids=edge_ids, report=report)
        rule_1.apply(node_id=new_node_id, edge_ids=other_edge_ids, report=report)    
    
    def find(self, report=False):
        anything_happened = False
        edges = self.infrastructure.edges
        for edge_ids in edges:
            for other_edge_ids in edges:
                if edge_ids != other_edge_ids:
                    result, intersection = self.check(edge_ids, other_edge_ids)
                    if result is True:
                        if report is True:
                            print(f"# {self.name}:")
                        self.apply(edge_ids, other_edge_ids, intersection, report)
                        # Inform an activity
                        anything_happened = True
                # Inform an activity     
                if anything_happened is True:
                    break
            # Inform an activity     
            if anything_happened is True:
                break
        return anything_happened
    

if __name__ == "__main__":

    from piperabm.infrastructure import Infrastructure

    infrastructure = Infrastructure()
    infrastructure.add_street(pos_1=[0, 0], pos_2=[10, 0])
    infrastructure.add_street(pos_1=[5, 5], pos_2=[5, -5])
    
    rule = Rule2(infrastructure, proximity_radius=1)
    rule.find(report=True)
    print(infrastructure)


from piperabm.model.graphics.style import *


node_radius = DEFAULT_NODE_RADIUS
font_size = DEFAULT_FONT_SIZE


infrastructure_node_style = {
    "home": {
        "color": "b",    
    "radius": node_radius,
    },
    "junction": {
        "color": "k",
        "radius": 0,
    },
    "market": {
        "color": "g",
        "radius": node_radius * 10,
    }
}

infrastructure_edge_style = {
    "street": {
        "color": "k",
    },
    "neighborhood_access": {
        "color": "silver",
    },
}

infrastructure_style = {
    "font": font_size,
    "node": infrastructure_node_style,
    "edge": infrastructure_edge_style,
}


if __name__ == "__main__":
    print(infrastructure_style)


from piperabm.tools.coordinate.projection.mercator import Mercator
from piperabm.tools.coordinate.projection.latlong_xyz import latlong_xyz, xyz_latlong
from piperabm.tools.coordinate import rotate


EARTH_RADIUS = 6378000 # meters


def latlong_xy(latitude_0=0, longitude_0=0, latitude=0, longitude=0):
    """
    Convert (latitude, logitude) to (x, y).

    :param latitude_0: latitude of the point which will be projected to (x=0, y=0).
    :param longitude_0: longitude of the point which will be projected to (x=0, y=0).
    :return: (x, y)
    """
    vector = latlong_xyz(latitude, longitude)
    vector = rotate.z(vector, longitude_0)
    vector = rotate.y(vector, -latitude_0)
    new_latitude, new_longitude = xyz_latlong(vector)
    x, y = Mercator.project(new_latitude, new_longitude, radius=EARTH_RADIUS)
    return x, y

def xy_latlong(latitude_0=0, longitude_0=0, x=0, y=0):
    """
    Convert (x, y) to (latitude, logitude).

    :param latitude_0: latitude of the point which will be projected to (x=0, y=0).
    :param longitude_0: longitude of the point which will be projected to (x=0, y=0).
    :return: (latitude, logitude)
    """
    new_latitude, new_longitude = Mercator.inverse(x, y, radius=EARTH_RADIUS)
    vector = latlong_xyz(new_latitude, new_longitude)
    vector = rotate.y(vector, latitude_0)
    vector = rotate.z(vector, -longitude_0)
    latitude, longitude = xyz_latlong(vector)
    return latitude, longitude


if __name__ == '__main__':
    latitude_0 = 70
    longitude_0 = -150

    latitude = latitude_0 + 1
    longitude = longitude_0 + 1

    x, y = latlong_xy(latitude_0, longitude_0, latitude, longitude)
    print(f'x, y: {x}, {y}')
    
    latitude, longitude = xy_latlong(latitude_0, longitude_0, x, y)
    print(f'latitude, longitude: {latitude}, {longitude}')


"""
Grid world without access to market
"""

import piperabm as pa


model = pa.Model(seed=2)
model.infrastructure.coeff_usage = 1
model.infrastructure.coeff_weather = 1
model.infrastructure.generate(
    homes_num=10,
    grid_size=[15, 10],
    grid_num=[6, 6],
    imperfection_percentage=10
)
model.set_seed(None)


if __name__ == "__main__":
    model.infrastructure.show()



from copy import deepcopy

from piperabm.infrastructure.samples.infrastructure_1 import model as model_infrastructure


model = deepcopy(model_infrastructure)
model.society.average_income = 1
agent_id = 1
home_id = model.infrastructure.homes[0]
model.society.add_agent(
    id=agent_id,
    home_id=home_id,
    socioeconomic_status=1,
    resources={
        "food": 1,
        "water": 1,
        "energy": 1,
    },
    balance=100
)


if __name__ == "__main__":
    destination_id = 2
    print(f"estimated_distance: {model.society.estimated_distance(agent_id, destination_id)} meters")
    print(f"estimated_duration: {model.society.estimated_duration(agent_id, destination_id)} seconds")
    #print(f"path: {model.society.path(agent_id, destination_id)}")
    #print(model.society.serialize())
    #model.show()


from copy import deepcopy

from piperabm.tools.print.serialized import Print


class Stay(Print):
    """
    Stay action
    """

    type = "stay"

    def __init__(self, action_queue, duration: float = 0):
        super().__init__()
        self.action_queue = action_queue  # Binding
        self.total_duration = duration
        self.elapsed = 0
        self.remaining = deepcopy(duration)
        self.done = False

    def update(self, duration: float, measure: bool = False):
        """
        Update status of action
        """
        if duration <= self.remaining:
            self.remaining -= duration
            self.elapsed += duration
            duration = 0
        else:
            duration -= self.remaining
            self.remaining = 0
            self.elapsed = self.total_duration
            self.done = True
        return duration
    
    def serialize(self) -> dict:
        """
        Serialize
        """
        data = {}
        data["type"] = self.type
        data["total_duration"] = self.total_duration
        data["remaining"] = self.remaining
        data["elapsed"] = self.elapsed
        data["done"] = self.done
        return data

    def deserialize(self, data: dict) -> None:
        """
        Deserialize
        """
        self.total_duration = data["total_duration"]
        self.elapsed = data["elapsed"]
        self.remaining = data["remaining"]
        self.done = data["done"] 


if __name__ == "__main__":

    from piperabm.society.samples.society_1 import model
    from piperabm.society.actions.action import Move

    agent_id = 1
    destination_id = 2
    action_queue = model.society.actions[agent_id]

    stay = Stay(
        action_queue=action_queue,
        duration=5
    )
    action_queue.add(stay)

    path = model.infrastructure.path(
        id_start=model.society.get_current_node(id=agent_id),
        id_end=destination_id
    )
    move = Move(
        action_queue=action_queue,
        path=path,
        usage=1
    )
    action_queue.add(move)

    print(f"time: {model.time}, pos: {model.society.get_pos(agent_id)}")

    model.update(duration=4)

    print(f"time: {model.time}, pos: {model.society.get_pos(agent_id)}")

    model.update(duration=14)

    print(f"time: {model.time}, pos: {model.society.get_pos(agent_id)}")


import numpy as np

from piperabm.society.actions.action import Move, Stay
from piperabm.tools.symbols import SYMBOLS


class DecisionMaking:
    """
    Methods related to agents' decision-making
    """

    def preasssumed_destinations(self, agent_id: int) -> list:
        """
        Return preassumed destinations (which are market nodes here)
        """
        result = []
        nodes = self.infrastructure.markets
        for node in nodes:
            path_exists = self.infrastructure.has_path(id_start=self.get_current_node(id=agent_id), id_end=node)
            if path_exists is True:
                result.append(
                    {
                        'id': node,
                        'score': self.destination_score(agent_id=agent_id, destination_id=node, is_market=True)
                    }
                )
        return result
    
    def search_destinations(self, agent_id: int) -> list:
        """
        Return possible search destinations which are NOT market nodes
        """
        result = []
        nodes = set()
        # Friends
        friends = self.ego(id=agent_id, type='friend')
        friends_homes = []
        for friend in friends:
            friends_homes.append(self.get_home_id(id=friend))
        nodes |= set(friends_homes)
        # Neighbors
        neighbors = self.ego(id=agent_id, type='neighbor')
        neighbors_homes = []
        for neighbor in neighbors:
            neighbors_homes.append(self.get_home_id(id=neighbor))
        nodes |= set(neighbors_homes)
        nodes = list(nodes)
        for node in nodes:
            path_exists = self.infrastructure.has_path(id_start=self.get_current_node(id=agent_id), id_end=node)
            if path_exists is True:
                result.append(
                    {
                        'id': node,
                        'score': self.destination_score(agent_id=agent_id, destination_id=node, is_market=False)
                    }
                )
        return result

    def destination_score(self, agent_id: int, destination_id: int, is_market: bool) -> float:
        """
        Destination score
        """
        # Calculate the estimated amount of fuel required
        travel_duration = self.estimated_duration(agent_id, destination_id)
        fuel_resources = {}
        for name in self.resource_names:
            fuel_resources[name] = self.transportation_resource_rates[name] * travel_duration
        # Calculate the value of required fuel
        fuel_possible = True
        for name in self.resource_names:
            if fuel_resources[name] > self.get_resource(id=agent_id, name=name):
                fuel_possible = False
        if fuel_possible is True:
            total_fuel_value = 0
            for name in self.resource_names:
                fuel_value = fuel_resources[name] * self.prices[name]
                total_fuel_value += fuel_value
        else:
            total_fuel_value = SYMBOLS['inf']
        # Calculate the value of resources there
        resources_there = self.resources_in(node_id=destination_id, is_market=is_market)
        total_value_there = 0
        for name in self.resource_names:
            total_value_there += resources_there[name] * self.prices[name]
        # Calculate score
        score = total_value_there - total_fuel_value
        return score
    
    def estimated_distance(self, agent_id: int, destination_id: int) -> float:
        """
        Estimated distance between agent and destination
        """
        return self.infrastructure.heuristic_paths.estimated_distance(
            id_start=self.get_current_node(id=agent_id),
            id_end=destination_id
        )

    def estimated_duration(self, agent_id, destination_id) -> float:
        """
        Estimated duration of reaching a certain destination
        """
        estimated_distance = self.estimated_distance(agent_id, destination_id)
        speed = self.speed
        return estimated_distance / speed

    def decide_destination(self, agent_id: int, duration: float) -> None:
        """
        Decide the destination
        """
        destination_id = None
        critical_stay_length = duration
        action_queue = self.get_action_queue(id=agent_id)
        # Find suitable market
        destinations = self.preasssumed_destinations(agent_id=agent_id)
        destinations = sorted(destinations, key=lambda x: x['score'], reverse=True)
        suitable_destination_found = False
        for destination in destinations:
            path = self.infrastructure.path(
                id_start=self.get_current_node(id=agent_id),
                id_end=destination['id']
            )
            move_go = Move(
                action_queue=action_queue,
                path=path,
                usage=self.transportation_degradation
            )
            # Stay (at the destination)
            stay_length = self.max_time_outside - (2 * move_go.total_duration)
            if stay_length > critical_stay_length:
                suitable_destination_found = True
                destination_id = destination['id']
                break
        if suitable_destination_found is True:
            self.go_and_comeback_and_stay(action_queue, move_go, stay_length)
        elif suitable_destination_found is False:
            destinations = self.search_destinations(agent_id=agent_id)
            destinations = sorted(destinations, key=lambda x: x['score'], reverse=True)
            for destination in destinations:
                path = self.infrastructure.path(
                    id_start=self.get_current_node(id=agent_id),
                    id_end=destination['id']
                )
                move_go = Move(
                    action_queue=action_queue,
                    path=path,
                    usage=self.transportation_degradation
                )
                # Stay (at the destination)
                stay_length = self.max_time_outside - (2 * move_go.total_duration)
                if stay_length > critical_stay_length:
                    suitable_destination_found = True
                    destination_id = destination['id']
                    break
            if suitable_destination_found is True:
                self.go_and_comeback_and_stay(action_queue, move_go, stay_length)
            elif suitable_destination_found is False:
                pass # No suitable destination found
        return destination_id

    def go_and_comeback_and_stay(self, action_queue, move_go, stay_length) -> None:
        """
        A complete daily cycle of choosing and going to a destination, waiting there for trade, and coming back home
        """
        # Move
        action_queue.add(move_go)
        stay_destination = Stay(
            action_queue=action_queue,
            duration=stay_length
        )
        # Stay
        action_queue.add(stay_destination)
        # Move
        move_back = move_go.reverse()
        action_queue.add(move_back) 
        # Stay
        stay_length = self.activity_cycle - action_queue.total_duration
        stay_home = Stay(
            action_queue=action_queue,
            duration=stay_length
        )
        action_queue.add(stay_home)


if __name__ == "__main__":

    from piperabm.infrastructure.samples.infrastructure_1 import model

    agent_id = 0
    home_id = 1
    destination_id = 2
    model.society.add_agent(
        home_id=home_id,
        id=agent_id,
    )

    print("preassumed destinations: ", model.society.preasssumed_destinations(agent_id=agent_id))
    print("search destinations: ", model.society.search_destinations(agent_id=agent_id))
    print("estimated distance: ", model.society.estimated_distance(agent_id=agent_id, destination_id=destination_id))
    print("estimated duration: ", model.society.estimated_duration(agent_id=agent_id, destination_id=destination_id))
    print("destination score: ", model.society.destination_score(agent_id=agent_id, destination_id=destination_id, is_market=True))
    print("destination id: ", model.society.decide_destination(agent_id=agent_id, duration=100))
    #print(model.society.get_action_queue(id=agent_id))



import os
import matplotlib.pyplot as plt
from copy import deepcopy

from piperabm.tools.average import average as avg


class Accessibility:
    """
    Manage accessibility measurement
    """

    type = 'accessibility'

    def __init__(self, measurement):
        self.measurement = measurement
        self.values = {}
        self.resource_names = ['food', 'water', 'energy']

    def add(self, id: int, value: float) -> None:
        """
        Add new accessibility value
        """
        if not id in self.values:
            self.values[id] = []
        self.values[id].append(value)

    def get(self, agent: int, resource: str, time_step: int):
        """
        Get the desired entry
        """
        return self.values[agent][time_step][resource]

    @property
    def agents(self):
        """
        Return a list of agents id
        """
        return list(self.values.keys())
    
    def rearrange(self, agents='all', resources='all', _from=None, _to=None):
        """
        Return accessibility values for all resources for across agents
        """
        if agents == 'all':
            ids = self.agents
        elif isinstance(agents, list):
            ids = agents
        elif isinstance(agents, int):
            ids = [agents]
        elif agents is None:
            ids = self.agents
        else:
            raise ValueError
        
        if resources == 'all':
            resource_names = self.resource_names
        elif isinstance(resources, list):
            resource_names = resources
        elif isinstance(resources, str) and \
        resources != 'all':
            resource_names = [resources]
        elif resources is None:
            resource_names = self.resource_names
        else:
            raise ValueError
        
        if _from is None:
            _from = 0

        if _to is None:
            _to = self.len
        
        result = {}
        for agent_id in ids:
            result[agent_id] = {}
            for resource_name in resource_names:
                values = []
                for i in range(_from, _to):
                    value = self.get(
                            agent=agent_id,
                            resource=resource_name,
                            time_step=i
                        )
                    values.append(value)
                result[agent_id][resource_name] = values

        return result
    
    @property
    def len(self) -> int:
        """
        Return total number of entries
        """
        return self.measurement.len

    def sum_resources(self, agents='all', resources='all', _from=None, _to=None) -> dict:
        """
        Sum resources
        """
        data = self.rearrange(agents=agents, resources=resources, _from=_from, _to=_to)
        if _from is None:
            _from = 0
        if _to is None:
            _to = self.len
        length = _to - _from
        result = {}
        for agent_id in data:
            result[agent_id] = []
            agent_data = data[agent_id]
            for i in range(length):
                vals = []
                for resource_name in agent_data:
                    acc_vals = agent_data[resource_name]
                    vals.append(acc_vals[i])
                result[agent_id].append(avg.geometric(values=vals))
        return result
    
    def sum_agents(self, agents='all', resources='all', _from=None, _to=None) -> list:
        """
        Sum agents
        """
        data = self.sum_resources(agents=agents, resources=resources, _from=_from, _to=_to)
        if _from is None:
            _from = 0
        if _to is None:
            _to = self.len
        length = _to - _from
        result = []
        for i in range(length):
            vals = []
            for agent_id in data:
                agent_data = data[agent_id]
                vals.append(agent_data[i])
            result.append(avg.arithmetic(values=vals))
        return result
    
    def __call__(self, agents='all', resources='all', _from=None, _to=None) -> list:
        return self.sum_agents(agents=agents, resources=resources, _from=_from, _to=_to)
    
    def average(self, agents='all', resources='all', _from=None, _to=None) -> float:
        """
        Calculate total average
        """
        values = self.__call__(agents=agents, resources=resources, _from=_from, _to=_to)
        weights = self.measurement.delta_times(_from=_from, _to=_to)
        result = avg.arithmetic(values=values, weights=weights)
        if isinstance(result, complex):
            result = float(result.real)
        return result
    
    def create_plot(self, agents='all', resources='all', _from=None, _to=None, info=None):
        """
        Create plot
        """
        fig, ax = plt.subplots()
        title = "Accessibility"
        ylabel = deepcopy(title)
        if info is not None:
            title += info
        ax.set_title(title)
        xs = self.measurement.filter_times(_from=_from, _to=_to)
        yx = self.__call__(agents=agents, resources=resources, _from=_from, _to=_to)
        ax.plot(xs, yx, color='blue')
        ax.set_xlabel("Time")
        ax.set_ylabel(ylabel)
        return fig

    def show(self, agents='all', resources='all', _from=None, _to=None, info=None):
        """
        Draw plot
        """
        fig = self.create_plot(
            agents=agents,
            resources=resources,
            _from=_from,
            _to=_to,
            info=info
        )
        plt.show()

    def save(self, agents='all', resources='all', _from=None, _to=None, info=None):
        """
        Save plot
        """
        fig = self.create_plot(
            agents=agents,
            resources=resources,
            _from=_from,
            _to=_to,
            info=info
        )
        path = self.measurement.result_directory
        filepath = os.path.join(path, self.type)
        fig.savefig(filepath)

    def serialize(self) -> dict:
        """
        Serialize
        """
        return {
            'values': self.values,
            'type': self.type
        }
    
    def deserialize(self, data: dict) -> None:
        """
        Deserialize
        """
        self.values = data['values']

    
if __name__ == "__main__":
    
    from piperabm.model.measurement import Measurement

    measure = Measurement()
    hour = 3600
    measure.add_time(0 * hour) # Base

    # 1
    measure.add_time(value=1*hour)
    measure.accessibility.add(id=1, value={'food': 1, 'water': 1, 'energy': 1})
    measure.accessibility.add(id=2, value={'food': 0.8, 'water': 0.7, 'energy': 0.6})
    # 2
    measure.add_time(value=2*hour)
    measure.accessibility.add(id=1, value={'food': 0.9, 'water': 0.8, 'energy': 0.7})
    measure.accessibility.add(id=2, value={'food': 0.5, 'water': 0.6, 'energy': 0.4})
    # 3
    measure.add_time(value=3*hour)
    measure.accessibility.add(id=1, value={'food': 0.8, 'water': 0.7, 'energy': 0.6})
    measure.accessibility.add(id=2, value={'food': 0.2, 'water': 0.4, 'energy': 0.3})
    # 4
    measure.add_time(value=4*hour)
    measure.accessibility.add(id=1, value={'food': 0.7, 'water': 0.6, 'energy': 0.5})
    measure.accessibility.add(id=2, value={'food': 0, 'water': 0.3, 'energy': 0.2})
    # 5
    measure.add_time(value=5*hour)
    measure.accessibility.add(id=1, value={'food': 0.6, 'water': 0.5, 'energy': 0.4})
    measure.accessibility.add(id=2, value={'food': 0, 'water': 0.3, 'energy': 0.2})

    agents = 'all'
    resources = 'all'
    _from = None
    _to = None
    print("accessibilities: ", measure.accessibility(agents=agents, resources=resources, _from=_from, _to=_to))
    print("average: ", measure.accessibility.average(agents=agents, resources=resources, _from=_from, _to=_to))
    measure.accessibility.show(agents=agents, resources=resources, _from=_from, _to=_to)



import numpy as np

from piperabm.tools.vector import vector as vc


def latlong_xyz(latitude_degree=0, longitude_degree=0, radius=1):
    """
    Convert geographic coordinates (latitude, longitude) to 3D Cartesian coordinates (x, y, z).

    :param latitude_degree: The latitude of the point in degrees.
    :param longitude_degree: The longitude of the point in degrees.
    :return: A numpy array representing a vector (x, y, z).
    """
    # Convert angles to radians
    latitude_rad = np.radians(latitude_degree)
    longitude_rad = np.radians(longitude_degree)

    # Spherical to Cartesian conversion
    x = radius * np.cos(latitude_rad) * np.cos(longitude_rad)
    y = radius * np.cos(latitude_rad) * np.sin(longitude_rad)
    z = radius * np.sin(latitude_rad)

    return np.array([x, y, z])


def xyz_latlong(vector):
    """
    Convert 3D Cartesian coordinates (x, y, z) to geographic coordinates (latitude, longitude).

    :param vector: A numpy array representing a vector (x, y, z).
    :return: A tuple (latitude, longitude) in degrees.
    """
    if isinstance(vector, list):
        vector = np.array(vector)

    # Normalize the vector to ensure it has a magnitude of 1
    x, y, z = vc.normalize(vector)

    # Calculate latitude and longitude
    latitude_degree = np.degrees(np.arcsin(z))
    longitude_degree = np.degrees(np.arctan2(y, x))

    return latitude_degree, longitude_degree


if __name__ == '__main__':
    latitude = 70
    longitude = -150
    vector = latlong_xyz(latitude, longitude)
    print(vector)
    latitude, longitude = xyz_latlong(vector)
    print(latitude, longitude)
    



from piperabm.tools.gini.gini_coefficient import gini_coefficient
from piperabm.tools.gini.gini_lognormal import GiniLogNormal


class gini:
    """
    A module to work with gini index
    """

    def coefficient(sample):
        """
        Calculate gini coefficient
        """
        return gini_coefficient(sample)
    
    def lognorm(gini_index: float = 0, average: float = 1):
        """
        Create a lognormal distribution by gini index and average
        """
        return GiniLogNormal(gini_index, average)
    

if __name__ == "__main__":
    
    from piperabm.tools.average import average as avg

    incomes = [100, 300, 500, 700, 900, 300, 500, 700, 500]
    gini_index = gini.coefficient(incomes)
    average = sum(incomes) / len(incomes)
    distribution = gini.lognorm(gini_index=gini_index, average=average)
    sample = distribution.rvs(sample_size=100, percision=0.02)
    print("Averages ratio: ", avg.arithmetic(values=sample) / avg.arithmetic(values=incomes))
    print("Gini index ratio: ", gini.coefficient(sample) / gini_index)


from piperabm.tools.nx_query import NxSet


class Set(NxSet):
    """
    Set attributes to network elements
    """

    def set_adjusted_length(self, ids: list, value: float) -> None:
        """
        Set edge *adjusted_length* value
        """
        self.set_edge_attribute(ids=ids, attribute='adjusted_length', value=value)
    
    def set_usage_impact(self, ids: list, value: float) -> None:
        """
        Set edge *usage_impact* value
        """
        self.set_edge_attribute(ids=ids, attribute='usage_impact', value=value)

    def set_climate_impact(self, ids: list, value: float) -> None:
        """
        Set edge *climate_impact* value
        """
        self.set_edge_attribute(ids=ids, attribute='climate_impact', value=value)

    def set_resource(self, id: int, name: str, value: float) -> None:
        """
        Set market *resource* value
        """
        if value <= 0:
            value = 0
        self.set_node_attribute(id=id, attribute=name, value=value)

    def set_balance(self, id: int, value: float) -> None:
        """
        Set market *balance* value
        """
        if value < 0:
            value = 0
        self.set_node_attribute(id=id, attribute='balance', value=value)


from piperabm.tools.print.stat import Print


class Stat(Print):
    """
    Stats of the network
    """
    
    @property
    def stat(self):
        """
        Return stats of the network
        """
        return {
            "node": {
                "alive": len(self.alives),
                "dead": len(self.deads),
                "total": len(self.agents),
            },
            "edge": {
                "family": len(self.families),
                "friend": len(self.friends),
                "neighbor": len(self.neighbors),
            },
        }
    

if __name__ == "__main__":

    from piperabm.model import Model

    model = Model()
    model.infrastructure.add_home(pos=[0, 0])
    model.bake()
    model.society.generate(num=1)

    print(model.society)


import numpy as np


def normalize(vector, ndarray=False):
    """
    Normalize a vector to have a magnitude of 1.
    
    :param v: A numpy array representing a vector.
    :return: Normalized vector.
    """
    result = None
    magnitude = np.linalg.norm(vector)
    if magnitude == 0:
        raise ValueError("Cannot normalize a zero vector")
    result = vector / magnitude
    if ndarray is False:
        result = [float(num) for num in result]  # Convert np.float64 to float explicitly
    return result


if __name__ == "__main__":
    vector = [1, 2 ,3]
    vector_normalized = normalize(vector)
    print(vector_normalized)


"""
Source:
https://stats.stackexchange.com/questions/286141/lognormal-parameters-knowing-gdp-per-capita-gini-coefficient-and-quintile-share
"""
import numpy as np
from scipy.special import erfinv
from scipy.stats import lognorm

from piperabm.tools.gini.gini_coefficient import gini_coefficient
from piperabm.tools.symbols import SYMBOLS


class GiniLogNormal:
    """
    Create a lognormal distribution by gini index and average
    """

    def __init__(self, gini_index: float = 0, average: float = 1):
        self.average = average
        if gini_index == 0:
            gini_index = SYMBOLS['eps']
        if gini_index < 0 or gini_index > 1:
            raise ValueError
        self.gini = gini_index

    @property
    def sigma(self):
        return 2 * erfinv(self.gini)
    
    @property
    def mu(self):
        return np.log(self.average) - (self.sigma ** 2) / 2
    
    @property
    def scale(self):
        return np.exp(self.mu)
    
    def rvs(self, sample_size: int = 1, percision: float = 0.03):
        """
        Generate random sample
        :sample_size: size of sample
        :percision: maximum difference between sample gini index and desired gini index
        """
        while True:
            result = lognorm(s=self.sigma, scale=self.scale).rvs(sample_size)
            if sample_size == 1:
                break
            else:
                result = [float(num) for num in result]  # Convert np.float64 to float explicitly
                diff = abs(gini_coefficient(result) - self.gini)
                if diff <= percision:
                    break
        return result


if __name__ == "__main__":
    distribution = GiniLogNormal(
        gini_index=0.45,
        average=1
    )
    print(distribution.mu, distribution.scale)
    print(distribution.sigma)
    sample = distribution.rvs(10)
    print(sample)



from copy import deepcopy

from piperabm.economy.trade import MultiResourceTrade
from piperabm.economy.trade.allocation import find_exchanges


class Trade:
    """
    Manage trade between players when updating
    """

    def trade(self, agents: list = [], markets: list = []):
        """
        Trade
        """
        names = self.resource_names
        players = []

        # Load agent players into solver
        for agent in agents:
            resources = {}
            enough_resources = {}
            for name in names:
                resources[name] = self.society.get_resource(id=agent, name=name)
                enough_resources[name] = self.society.get_enough_resource(id=agent, name=name)
            player = {
                'id': agent,
                'type': 'agent',
                'resources': resources,
                'enough_resources': enough_resources,
                'balance': self.society.get_balance(id=agent),
            }
            players.append(player)

        # Load market players into solver
        for market in markets:
            resources = {}
            enough_resources = {}
            for name in names:
                resources[name] = self.infrastructure.get_resource(id=market, name=name)
                enough_resources[name] = self.infrastructure.get_enough_resource(id=market, name=name)
            player = {
                'id': market,
                'type': 'market',
                'resources': resources,
                'enough_resources': enough_resources,
                'balance': self.infrastructure.get_balance(id=market)
            }
            players.append(player)

        # Solve
        players_initial = deepcopy(players)
        players = MultiResourceTrade.solve(players=players, prices=self.prices)
        transactions = find_exchanges(players_initial=players_initial, players_final=players)

        # Update values
        for player in players:
            # Update agent players
            if player['type'] == 'agent':
                for name in names:
                    self.society.set_resource(
                        id=player['id'],
                        name=name,
                        value=player['resources'][name]
                    )
                self.society.set_balance(
                    id=player['id'],
                    value=player['balance']
                )
            # Update market players
            elif player['type'] == 'market':
                for name in names:
                    self.infrastructure.set_resource(
                        id=player['id'],
                        name=name,
                        value=player['resources'][name]
                    )
                self.infrastructure.set_balance(
                    id=player['id'],
                    value=player['balance']
                )
        #print(transactions)
        return flatten_transactions(transactions)
    

def flatten_transactions(transactions):
    results = []
    for resource_name in transactions:
        for transaction in transactions[resource_name]:
            result = [
                transaction[0], # from
                transaction[1], # to
                transaction[2], # amount
                resource_name, # resource name
            ]
            results.append(result)
    #print(results)
    return results
    

if __name__ == "__main__":
    
    from piperabm.society.samples.society_0 import model

    model.society.average_income = 0
    agents = model.society.agents

    wealth_0 = model.society.wealth(agents[0])
    wealth_1 = model.society.wealth(agents[1])
    if wealth_0 < wealth_1:
        id_low = agents[0]  # Agent with lower wealth
        id_high = agents[1]  # Agent with higher wealth
    food = model.society.get_resource(id_low, 'food')
    model.society.set_resource(id_low, 'food', value=food/10)
    water = model.society.get_resource(id_low, 'water')
    model.society.set_resource(id_low, 'water', value=water/5)

    print("\n" + ">>> Initial:")
    for agent in agents:
        balance = model.society.get_balance(agent)
        food = model.society.get_resource(agent, 'food')
        water = model.society.get_resource(agent, 'water')
        energy = model.society.get_resource(agent, 'energy')
        print(f"id: {agent}, balance: {balance}, food: {food}, water: {water}, energy: {energy}")

    transactions = model.update(duration=1)

    print("\n" + ">>> " + "Exchanges: ")
    for transaction in transactions:
        print(f"from: {transaction[0]}, to: {transaction[1]}, amount: {transaction[2]}, resource: {transaction[3]}")

    print("\n" + ">>> Final:")
    for agent in agents:
        balance = model.society.get_balance(agent)
        food = model.society.get_resource(agent, 'food')
        water = model.society.get_resource(agent, 'water')
        energy = model.society.get_resource(agent, 'energy')
        print(f"id: {agent}, balance: {balance}, food: {food}, water: {water}, energy: {energy}")


import unittest


def run():
    loader = unittest.TestLoader()
    suite = loader.discover(start_dir="tests", pattern="test_*.py")
    runner = unittest.TextTestRunner()
    runner.run(suite)


if __name__ == "__main__":
    run()



import numpy as np


class Generate:
    """
    Generate infrastructure
    """
    
    def generate(
        self,
        homes_num: int = 1,
        grid_size: list = [1, 1],
        grid_num: list = [2, 2],
        imperfection_percentage: float = 0
    ):
        """
        Generate a grid world model
        """

        x_size = grid_size[0] * (grid_num[0] - 1)
        y_size = grid_size[1] * (grid_num[1] - 1)
        x_range = grid_size[0] * grid_num[0]
        y_range = grid_size[1] * grid_num[1]

        # Streets
        for i in range(grid_num[0]):
            x = (i * grid_size[0]) - (x_size / 2)
            self.add_street(
                pos_1=[x, 0 - (y_size / 2)],
                pos_2=[x, y_size - (y_size / 2)]
            )
        for j in range(grid_num[1]):
            y = (j * grid_size[1]) - (y_size / 2)
            self.add_street(
                pos_1=[0 - (x_size / 2), y],
                pos_2=[x_size - (x_size / 2), y]
            )

        # Homes
        def generate_random_point(x_range, y_range):
            x = float(np.random.uniform(-x_range/2, x_range/2))
            y = float(np.random.uniform(-y_range/2, y_range/2))
            pos = [x, y]
            #pos = [float(num) for num in pos]  # Convert type from np.float64 to float
            return pos

        for i in range(homes_num):
            self.add_home(pos=generate_random_point(x_range, y_range))

        self.bake()

        # Random impact
        edges = self.random_edges(percent=imperfection_percentage)
        self.impact(edges=edges)


if __name__ == "__main__":

    from piperabm.infrastructure import Infrastructure

    infrastructure = Infrastructure()
    infrastructure.generate(
        homes_num=10,
        grid_size=[15, 10],
        grid_num=[6, 6],
    )
    infrastructure.show()


"""
Simple world
"""

import piperabm as pa


model = pa.Model()
model.set_seed(2)
model.infrastructure.coeff_usage = 1
model.infrastructure.coeff_weather = 1
model.infrastructure.add_market(
    pos=[40, -40],
    name='market',
    id=0,
    resources={'food': 100, 'water': 100, 'energy': 100}
)
model.infrastructure.add_home(pos=[-60, 40], name='home 1', id=1)
model.infrastructure.add_home(pos=[200, 20], name='home 2', id=2)
model.infrastructure.add_home(pos=[100, -180], name='home 3', id=3)
model.infrastructure.add_street(pos_1=[-60, 40], pos_2=[0, 0], name='street 1')
model.infrastructure.add_street(pos_1=[0, 0], pos_2=[80, 60], name='street 2')
model.infrastructure.add_street(pos_1=[80, 60], pos_2=[200, 20], name='street 3')
model.infrastructure.add_street(pos_1=[0, 0], pos_2=[100, -180], name='street 4')
model.infrastructure.bake()
model.set_seed(None)


if __name__ == "__main__":
    model.infrastructure.show()



import networkx as nx

from piperabm.tools.coordinate import distance as ds
from piperabm.tools.nx_serializer import nx_serialize, nx_deserialize


class HeuristicPaths:
    """
    A network to capture the Eucledean distance between all nodes
    """

    type = "heuristic paths"

    def __init__(self):
        self.G = nx.Graph()  # Heuristic

    def estimated_distance(self, id_start, id_end):
        """
        Return estimated distance between two nodes
        """
        edge = self.G.edges[id_start, id_end]
        result = edge['distance']
        return result

    def create(self, infrastructure):
        """
        Create graph
        """
        self.G = nx.Graph()  # Reset
        nodes = infrastructure.nodes
        for id_1 in nodes:
            for id_2 in nodes:
                if id_1 == id_2:
                    distance = 0
                else:
                    distance = ds.point_to_point(
                        point_1=infrastructure.get_pos(id_1),
                        point_2=infrastructure.get_pos(id_2)
                    )
                self.G.add_edge(id_1, id_2, distance=distance)
    
    def serialize(self):
        """
        Serialize
        """
        dictionary = {
            'G': nx_serialize(self.G),
            'type': self.type,
        }
        return dictionary
    
    def deserialize(self, dictionary):
        """
        Deserialize
        """
        self.G = nx_deserialize(dictionary['G'])

    def __str__(self):
        return self.G.__str__()


if __name__ == "__main__":

    from piperabm.infrastructure import Infrastructure

    infrastructure = Infrastructure()
    infrastructure.add_street(pos_1=[0, 0], pos_2=[10, 0])
    infrastructure.add_home(pos=[0, 0], id=1)
    infrastructure.add_home(pos=[10, 0], id=2)
    infrastructure.bake()

    # Test elements count
    junctions = infrastructure.stat['node']['junction']
    homes = infrastructure.stat['node']['home']
    total_nodes = junctions + homes
    print("Expected nodes: ", total_nodes)
    print("Expected edges: ", (total_nodes + 1) * (total_nodes / 2))
    print("Resulting nodes and edges: ", infrastructure.heuristic_paths)

    # Test functionality
    print(infrastructure.heuristic_paths.estimated_distance(id_start=1, id_end=2))


import os
from copy import deepcopy

import piperabm as pa
from piperabm.infrastructure.samples.infrastructure_3 import model as model_3


path = os.path.dirname(os.path.realpath(__file__))

# Setup
model = deepcopy(model_3)
model.path = path
model.society.neighbor_radius = 270
model.society.generate(
    num=10,
    gini_index=0.45,
    average_balance=100
)

# Run
model.run(n=50, save=True, resume=False, report=True, step_size=10)

# Measure
measurement = pa.Measurement(path)
measurement.measure(resume=False, report=True)


import networkx as nx

from piperabm.society.query import Query
from piperabm.society.generate import Generate
from piperabm.society.decision_making import DecisionMaking
from piperabm.society.update import Update
from piperabm.society.serialize import Serialize
from piperabm.society.graphics import Graphics
from piperabm.society.stat import Stat
from piperabm.tools.gini import gini
from piperabm.economy import accessibility
from piperabm.society.info import *


class Society(
    Query,
    Generate,
    DecisionMaking,
    Update,
    Serialize,
    Graphics,
    Stat
):
    """
    Represent society network
    """

    type = "society"

    def __init__(
            self,
            average_income: float = 1000,
            neighbor_radius: float = 0,
            max_time_outside: float = 8 * (60 * 60), # 8 hours
            activity_cycle: float = 24 * (60 * 60), # 24 hours
            transportation_resource_rates: dict = transportation_resource_rates,
            idle_resource_rates: dict = idle_resource_rates,
            speed: float = speed,
            transportation_degradation: float = 1,
        ):
        self.G = nx.MultiGraph()
        self.model = None # Binding
        self.actions = {}
        self.average_income = average_income
        self.neighbor_radius = neighbor_radius
        if max_time_outside > activity_cycle:
            raise ValueError('"max_time_outside" should be less than "activity_cycle"')
        self.max_time_outside = max_time_outside
        self.activity_cycle = activity_cycle
        self.idle_resource_rates = idle_resource_rates
        self.transportation_resource_rates = transportation_resource_rates
        self.speed = speed
        self.transportation_degradation = transportation_degradation

    @property
    def infrastructure(self):
        """
        Alias
        """
        if self.model is not None:
            result = self.model.infrastructure
        else:
            result = None
        return result
    
    @property
    def resource_names(self) -> list:
        """
        Alias
        """
        return self.model.resource_names
    
    @property
    def prices(self) -> dict:
        """
        Alias
        """
        return self.model.prices

    @property
    def gini_index(self) -> float:
        """
        Calculate the current gini index of society
        """
        data = []
        for id in self.agents:
            data.append(self.wealth(id))
        return gini.coefficient(data)

    def accessibility_resource(self, id: int, name: str):
        """
        Calculate accessibility for certain resource
        """
        amount = self.get_resource(id=id, name=name)
        enough_amount = self.get_enough_resource(id=id, name=name)
        return accessibility(resource=amount, enough_resource=enough_amount)
    
    def accessibility(self, id: int) -> dict:
        """
        Calculate accessibility for all resources
        """
        result = {}
        for name in self.resource_names:
            result[name] = self.accessibility_resource(id=id, name=name)
        return result
    

if __name__ == "__main__":

    from piperabm.model import Model

    model = Model()
    model.infrastructure.add_home(pos=[0, 0])
    model.bake()
    model.society.generate(num=1)

    print(model.society)


class Interaction:
    """
    Manage interactions graph
    """

    type = 'interaction'

    def __init__(self, measurement):
        self.measurement = measurement
        self.values = {}

    def add(self, id_from: int, id_to: int, resource_name: str, resource_amount: float) -> None:
        """
        Add new accessibility value
        """
        if not id_from in self.values:
            self.values[id_from] = {}
        if not id_to in self.values[id_from]:
            self.values[id_from][id_to] = {}
        if not resource_name in self.values[id_from][id_to]:
            self.values[id_from][id_to][resource_name] = []
        self.values[id_from][id_to][resource_name].append(resource_amount)

    def serialize(self) -> dict:
        return {
            'values': self.values,
            'type': self.type
        }
    
    def deserialize(self, data: dict) -> None:
        """
        Deserialize
        """
        self.values = data['values']


if __name__ == "__main__":
    
    from piperabm.model.measurement import Measurement

    measure = Measurement()
    hour = 3600
    measure.add_time(0 * hour) # Base

    # 1
    measure.add_time(value=1*hour)
    measure.accessibility.add(id=1, value={'food': 1, 'water': 1, 'energy': 1})
    measure.accessibility.add(id=2, value={'food': 0.8, 'water': 0.7, 'energy': 0.6})
    # 2
    measure.add_time(value=2*hour)
    measure.accessibility.add(id=1, value={'food': 0.9, 'water': 0.8, 'energy': 0.7})
    measure.accessibility.add(id=2, value={'food': 0.5, 'water': 0.6, 'energy': 0.4})
    # 3
    measure.add_time(value=3*hour)
    measure.accessibility.add(id=1, value={'food': 0.8, 'water': 0.7, 'energy': 0.6})
    measure.accessibility.add(id=2, value={'food': 0.2, 'water': 0.4, 'energy': 0.3})
    # 4
    measure.add_time(value=4*hour)
    measure.accessibility.add(id=1, value={'food': 0.7, 'water': 0.6, 'energy': 0.5})
    measure.accessibility.add(id=2, value={'food': 0, 'water': 0.3, 'energy': 0.2})
    # 5
    measure.add_time(value=5*hour)
    measure.accessibility.add(id=1, value={'food': 0.6, 'water': 0.5, 'energy': 0.4})
    measure.accessibility.add(id=2, value={'food': 0, 'water': 0.3, 'energy': 0.2})

    agents = 'all'
    resources = 'all'
    _from = None
    _to = None
    print("accessibilities: ", measure.accessibility(agents=agents, resources=resources, _from=_from, _to=_to))
    print("average: ", measure.accessibility.average(agents=agents, resources=resources, _from=_from, _to=_to))
    measure.accessibility.show(agents=agents, resources=resources, _from=_from, _to=_to)


import os

from piperabm.model.measurement import Measurement


path = os.path.dirname(os.path.realpath(__file__))

measure = Measurement(path=path, name='model')
hour = 3600
measure.add_time(0 * hour) # Base

# 1
measure.add_time(value=1*hour)
measure.add_accessibility(id=1, value={'food': 1, 'water': 1, 'energy': 1})
measure.add_accessibility(id=2, value={'food': 0.8, 'water': 0.7, 'energy': 0.6})
measure.add_travel_distance(value=1.1)
# 2
measure.add_time(value=2*hour)
measure.add_accessibility(id=1, value={'food': 0.9, 'water': 0.8, 'energy': 0.7})
measure.add_accessibility(id=2, value={'food': 0.5, 'water': 0.6, 'energy': 0.4})
measure.add_travel_distance(value=0.9)
# 3
measure.add_time(value=3*hour)
measure.add_accessibility(id=1, value={'food': 0.8, 'water': 0.7, 'energy': 0.6})
measure.add_accessibility(id=2, value={'food': 0.2, 'water': 0.4, 'energy': 0.3})
measure.add_travel_distance(value=0.3)
# 4
measure.add_time(value=4*hour)
measure.add_accessibility(id=1, value={'food': 0.7, 'water': 0.6, 'energy': 0.5})
measure.add_accessibility(id=2, value={'food': 0, 'water': 0.3, 'energy': 0.2})
measure.add_travel_distance(value=0.46)
# 5
measure.add_time(value=5*hour)
measure.add_accessibility(id=1, value={'food': 0.6, 'water': 0.5, 'energy': 0.4})
measure.add_accessibility(id=2, value={'food': 0, 'water': 0.3, 'energy': 0.2})
measure.add_travel_distance(value=0.2)

measure.save()


"""
Solves multi-resources for multiple players at fixed price
"""

from piperabm.economy.trade.nash_bargaining import NashBargaining as nb


class MultiResourceTrade:

    def resource_names(players: list):
        return players[0]['resources'].keys()

    def balance_allocations(players: list, prices: dict) -> list:
        # Balance allocation
        resource_names = MultiResourceTrade.resource_names(players=players)
        result = []
        for player in players:
            balance_allocation = {}
            needs = {}
            for resource_name in resource_names:
                need = max(0, (player['enough_resources'][resource_name] - player['resources'][resource_name]) * prices[resource_name])
                needs[resource_name] = need
            all_needs = sum(needs.values())
            for resource_name in resource_names:
                if all_needs != 0:
                    balance_allocation[resource_name] = needs[resource_name] / all_needs
                else:
                    balance_allocation[resource_name] = 0
            result.append(balance_allocation)
        return result
    
    def prepare(players: list, balance_allocations: list, resource_name: str):
        """
        Extract the player info
        """
        result = []
        for i, player in enumerate(players):
            new = {}
            new['id'] = player['id']
            new['resource'] = player['resources'][resource_name]
            new['enough_resource'] = player['enough_resources'][resource_name]
            new['balance'] = player['balance'] * balance_allocations[i][resource_name]
            result.append(new)
        return result
    '''
    def transactions(players: list, prices: dict):
        balance_allocations = MultiResourceTrade.balance_allocations(players)
        resource_names = MultiResourceTrade.resource_names(players=players)
        all_transactions = {}
        for resource_name in resource_names:
            market_players = MultiResourceTrade.prepare(
                players=players,
                balance_allocations=balance_allocations,
                resource_name=resource_name
            )
            transactions = nb.transactions(players=market_players, price=prices[resource_name])
            all_transactions[resource_name] = transactions
            #players = nb.apply(players, transactions)
        return all_transactions
    '''

    def sorted_markets(players, prices):
        """
        Sorted name of markets based on their size
        """
        market_sizes = {}
        for resource_name in MultiResourceTrade.resource_names(players):
            market_sizes[resource_name] = 0
        for player in players:
            resources = player['resources']
            for resource_name in resources:
                market_sizes[resource_name] += resources[resource_name]
        for resource_name in market_sizes:
            market_sizes[resource_name] *= prices[resource_name]
        return sorted(market_sizes, key=market_sizes.get, reverse=True)
    
    def transactions(players: list, prices: dict) -> dict:
        """
        Calculate the proper transactions
        """
        markets = MultiResourceTrade.sorted_markets(players=players, prices=prices)
        all_transactions = {}
        for market in markets:
            balance_allocations = MultiResourceTrade.balance_allocations(players=players, prices=prices)
            market_players = MultiResourceTrade.prepare(
                players=players,
                balance_allocations=balance_allocations,
                resource_name=market
            )
            transactions = nb.transactions(players=market_players, price=prices[market])
            all_transactions[market] = transactions
        return all_transactions
    
    def apply(players, transactions) -> list:
        """
        Apply transactions to players
        """
        num_players = len(players)
        resource_names = MultiResourceTrade.resource_names(players=players)
        for resource_name in resource_names:
            resource_transactions = transactions[resource_name]
            for i in range(num_players):
                players[i]['balance'] = players[i]['balance'] + resource_transactions['money'][i]
                players[i]['resources'][resource_name] = players[i]['resources'][resource_name] + resource_transactions['resource'][i]
        return players

    def solve(players: list, prices: dict) -> list:
        """
        Solve until convergence
        """
        i = 0
        #max_itteration = 1
        max_itteration = len(players) ** 2
        while i < max_itteration:
            transactions = MultiResourceTrade.transactions(players=players, prices=prices)
            #print(transactions)
            if MultiResourceTrade.check_empty(transactions) is True:
                break
            else:
                players = MultiResourceTrade.apply(players=players, transactions=transactions)
            i += 1
        '''
        transactions = MultiResourceTrade.transactions(players=players, prices=prices)
        players = MultiResourceTrade.apply(players=players, transactions=transactions)
        '''
        return players

    def check_empty(transactions: dict) -> bool:
        result = None
        threashold = 0
        transfers = 0
        for resource_name in transactions:
            transfers += sum(transactions[resource_name]['money'])
        if transfers <= threashold:
            result = True
        else:
            result = False
        return result


if __name__ == "__main__":
    prices = {
        'food': 10,
        'water': 10,
        'energy': 10
    }
    player_1 = {
        'id': 1,
        'type': 'agent',
        'resources': {
            'food': 9,
            'water': 2,
            'energy': 3
        },
        'enough_resources': {
            'food': 10,
            'water': 10,
            'energy': 10
        },
        'balance': 100,
    }
    player_2 = {
        'id': 2,
        'type': 'agent',
        'resources': {
            'food': 10,
            'water': 3,
            'energy': 10
        },
        'enough_resources': {
            'food': 10,
            'water': 10,
            'energy': 10
        },
        'balance': 100,
    }
    player_3 = {
        'id': 3,
        'type': 'agent',
        'resources': {
            'food': 3,
            'water': 10,
            'energy': 7
        },
        'enough_resources': {
            'food': 10,
            'water': 10,
            'energy': 10
        },
        'balance': 10,
    }
    players = [player_1, player_2, player_3]

    print("Solves multi-resources for multiple players at fixed price.")

    # Initial
    print("\n" + ">>> " + "Initial: ")
    for player in players:
        print(player['resources'], ', balance:', player['balance'])

    # Solve
    players = MultiResourceTrade.solve(players=players, prices=prices)

    # Final
    print("\n" + ">>> " + "Final: ")
    for player in players:
        print(player['resources'], ', balance:', player['balance'])


from piperabm.economy.trade.nash_bargaining import NashBargaining as nb


price = 10
player_1 = {
    'id': 1,
    'type': 'agent',
    'resource': 2,
    'enough_resource': 10,
    'balance': 100,
}
player_2 = {
    'id': 2,
    'type': 'agent',
    'resource': 8,
    'enough_resource': 10,
    'balance': 100,
}
players = [player_1, player_2]

transactions = nb.transactions(players, price)
players = nb.apply(players, transactions)


if __name__ == "__main__":
    #print(transactions)
    for player in players:
        print(f"resource: {player['resource']}, balance: {player['balance']}")



import math


def geometric(values: list, weights: list = None):
    result = None
    if weights is None:
        total = 1
        for value in values:
            total *= value
        result = total ** (1 / len(values))
    else:
        if len(values) != len(weights):
            raise ValueError("The lengths of values and weights must be the same.")
        
        if any(v <= 0 for v in values):
            raise ValueError("All values must be greater than zero.")
        
        log_values = [math.log(v) * w for v, w in zip(values, weights)]
        total_weight = sum(weights)
        
        if total_weight == 0:
            raise ValueError("The sum of weights must not be zero.")
        
        log_weighted_mean = sum(log_values) / total_weight
        result = math.exp(log_weighted_mean)
    return result


if __name__ == "__main__":
    values = [1, 2, 3]
    print(geometric(values=values))

    weights = [1, 1, 1]
    print(geometric(values=values, weights=weights))

    weights = [100, 1, 1]
    print(geometric(values=values, weights=weights))



from piperabm.tools.nx_query import NxGet


class Get(NxGet):
    """
    Get attributes from network elements
    """
    
    def get_pos(self, id: int):
        """
        Get node position
        """
        return [self.get_node_attribute(id, 'x', None), self.get_node_attribute(id, 'y', None)]
    
    def get_node_type(self, id: int) -> str:
        """
        Get node type
        """
        return self.get_node_attribute(id=id, attribute='type')

    def get_edge_type(self, ids: list) -> str:
        """
        Get edge type
        """
        return self.get_edge_attribute(ids=ids, attribute='type')    

    def get_node_name(self, id: int) -> str:
        """
        Get node name
        """
        return self.get_node_attribute(id=id, attribute='name')
    
    def get_edge_name(self, ids: list) -> str:
        """
        Get edge name
        """
        return self.get_edge_attribute(ids=ids, attribute='name')
    
    def get_length(self, ids: list) -> float:
        """
        Get edges length
        """
        return self.get_edge_attribute(ids=ids, attribute='length')
    
    def get_adjusted_length(self, ids: list) -> float:
        """
        Get edges *adjusted_length*
        """
        return self.get_edge_attribute(ids=ids, attribute='adjusted_length')
    
    def get_usage_impact(self, ids: list) -> float:
        """
        Get edges *usage_impact*
        """
        return self.get_edge_attribute(ids=ids, attribute='usage_impact')
    
    def get_climate_impact(self, ids: list) -> float:
        """
        Get edges *climate_impact*
        """
        return self.get_edge_attribute(ids=ids, attribute='climate_impact')
    
    def get_resource(self, id: int, name: str) -> float:
        """
        Get market *resource* value
        """
        return self.get_node_attribute(id=id, attribute=name)
    
    def get_enough_resource(self, id: int, name: str) -> float:
        """
        Get market *enough_resource* value
        """
        attribute = 'enough_' + name
        return self.get_node_attribute(id=id, attribute=attribute)
    
    def get_balance(self, id: int) -> float:
        """
        Get market *balance* value
        """
        return self.get_node_attribute(id=id, attribute='balance')



from piperabm.tools.vector.magnitude import magnitude
from piperabm.tools.vector.normalize import normalize


class vector:
    """
    A module to work with vectors
    """
    
    @staticmethod
    def magnitude(vector):
        return magnitude(vector)
    
    @staticmethod
    def normalize(vector, ndarray=False):
        return normalize(vector, ndarray)
    

if __name__ == "__main__":
    v = [1, 2 ,3]
    v_normalized = vector.normalize(v)
    v_normalized_magnitude = vector.magnitude(v_normalized)
    print(v_normalized_magnitude)


import numpy as np
from copy import deepcopy

from piperabm.society.actions.action_queue import ActionQueue
from piperabm.society.info import *


class Add:
    """
    Add new network elements
    """

    def check_id(self, id):
        """
        Check whether id already exists
        """
        if id is None:
            id = self.new_id()
        else:
            if id in self.agents:
                id = self.new_id()
                print("id already exists. replaced with new id.")
        return id

    def new_id(self) -> int:
        """
        Generate new unique random id
        """
        return int(np.random.randint(low=0, high=np.iinfo(np.int64).max, dtype=np.int64))
    
    def add_agent(
        self,
        home_id: int = None,
        id: int = None,
        name: str = '',
        socioeconomic_status: float = 1,
        resources: dict = {
            'food': 10,
            'water': 10,
            'energy': 10,
        },
        enough_resources: dict = None,
        balance: float = 0
    ):
        """
        Add agent node
        """
        if self.infrastructure.baked is False:
            raise ValueError("Model is not baked.")
        type = 'agent'
        id = self.check_id(id)
        self.actions[id] = ActionQueue(agent_id=id)
        self.actions[id].society = self # Binding
        if home_id is None:
            homes_id = self.infrastructure.homes
            home_id = int(np.random.choice(homes_id))
        pos = self.infrastructure.get_pos(id=home_id)
        resource_kwargs = {}
        if enough_resources is None:
            enough_resources = {}
            for resource_name in self.resource_names:
                enough_resources[resource_name] = None
        for resource_name in self.resource_names:
            if enough_resources[resource_name] is None:
                enough_resources[resource_name] = deepcopy(resources[resource_name])
            resource_kwargs[resource_name] = resources[resource_name]
            resource_kwargs['enough_'+resource_name] = enough_resources[resource_name]

        self.G.add_node(
            id,
            name=name,
            type=type,
            socioeconomic_status=socioeconomic_status,
            home_id=home_id,
            current_node=deepcopy(home_id),
            x=float(pos[0]),
            y=float(pos[1]),
            balance=balance,
            alive=True,
            **resource_kwargs
        )

        # Add relationship edge(s)
        # Family
        family_members = self.agents_from(home_id=home_id)
        for member in family_members:
            if member != id:
                self.add_family(id_1=id, id_2=member)
        
        # Neighbor
        neighbor_homes = self.infrastructure.nodes_closer_than(
            id=home_id,
            search_radius=self.neighbor_radius,
            nodes=self.infrastructure.homes,
            include_self=False
        )
        for neighbor_home_id in neighbor_homes:
            neighbors = self.agents_from(home_id=neighbor_home_id)
            for neighbor in neighbors:
                self.add_neighbor(id_1=id, id_2=neighbor)
        
    def add_family(
        self,
        id_1: int,
        id_2: int
    ):
        """
        Add family edge
        """
        type = 'family'
        home_id_1 = self.get_node_attribute(id=id_1, attribute='home_id')
        home_id_2 = self.get_node_attribute(id=id_1, attribute='home_id')
        if home_id_1 == home_id_2 and \
        id_1 != id_2:
            home_id = home_id_1
            self.G.add_edge(
                id_1,
                id_2,
                type=type,
                home_id=home_id
            )

    def add_friend(
        self,
        id_1: int,
        id_2: int
    ):
        """
        Add friend edge
        """
        type = 'friend'
        self.G.add_edge(
            id_1,
            id_2,
            type=type
        )

    def add_neighbor(
        self,
        id_1,
        id_2
    ):
        """
        Add neighbor edge
        """
        type = 'neighbor'
        home_id_1 = self.get_node_attribute(id=id_1, attribute='home_id')
        home_id_2 = self.get_node_attribute(id=id_2, attribute='home_id')
        if home_id_1 != home_id_2 and \
        id_1 != id_2:
            self.G.add_edge(
                id_1,
                id_2,
                type=type,
            )


if __name__ == "__main__":

    from piperabm.infrastructure.samples.infrastructure_0 import model

    model.society.add_agent()
    print(model.society)



import numpy as np


def line_line(
    line_1_point_1: list,
    line_1_point_2: list,
    line_2_point_1: list,
    line_2_point_2: list
):
    """
    Calculate intersecting point between two line segments
    Source: Computer Graphics by F.S. Hill
    """

    def perp(a):
        b = np.empty_like(a)
        b[0] = -a[1]
        b[1] = a[0]
        return b
    
    intersection = None

    line_1_point_1 = np.array(line_1_point_1)
    line_1_point_2 = np.array(line_1_point_2)
    line_2_point_1 = np.array(line_2_point_1)
    line_2_point_2 = np.array(line_2_point_2)

    delta_line_1 = line_1_point_2 - line_1_point_1
    delta_line_2 = line_2_point_2 - line_2_point_1
    delta_p = line_1_point_1 - line_2_point_1
    delta_line_1_perp = perp(delta_line_1)

    denominator = np.dot(delta_line_1_perp, delta_line_2)
    numerator = np.dot(delta_line_1_perp, delta_p)

    if denominator != 0: # check for parallel lines
        intersection = (numerator / denominator.astype(float)) * delta_line_2 + line_2_point_1
        intersection = list(intersection)

    return intersection


if __name__ == "__main__":
    line_1_point_1 = [1, 0]
    line_1_point_2 = [1, 2]
    line_2_point_1 = [0, 1]
    line_2_point_2 = [2, 1]
    point = line_line(line_1_point_1, line_1_point_2, line_2_point_1, line_2_point_2)
    print(point)


import networkx as nx

from piperabm.infrastructure.query import Query
from piperabm.infrastructure.generate import Generate
from piperabm.infrastructure.degradation import Degradation
from piperabm.infrastructure.path import Path
from piperabm.infrastructure.update import Update
from piperabm.infrastructure.serialize import Serialize
from piperabm.infrastructure.graphics import Graphics
from piperabm.infrastructure.stat import Stat
from piperabm.infrastructure.grammar import Grammar
from piperabm.infrastructure.heuristic_paths import HeuristicPaths


class Infrastructure(
    Query,
    Generate,
    Degradation,
    Path,
    Update,
    Serialize,
    Graphics,
    Stat
):
    """
    Represent infrastructure network
    """

    type = 'infrastructure'

    def __init__(
            self,
            coeff_usage: float = 0,
            coeff_weather: float = 0
        ):
        self.G = nx.Graph()
        self.model = None # Binding
        self.coeff_usage = coeff_usage
        self.coeff_weather = coeff_weather
        self.baked_streets = True
        self.baked_neighborhood = True
        self.heuristic_paths = HeuristicPaths()

    @property
    def resource_names(self) -> list:
        """
        Alias
        """
        return self.model.resource_names
    
    @property
    def prices(self) -> dict:
        """
        Alias
        """
        return self.model.prices

    @property
    def baked(self) -> bool:
        """
        Check if the network is fully baked
        """
        result = False
        if self.baked_streets is True and \
        self.baked_neighborhood is True:
            result = True
        return result
    
    def bake(
            self,
            report: bool = False,
            proximity_radius: float = 1,
            search_radius: float = None
        ):
        """
        Bake the network using grammar rules
        """
        if self.baked is False:
            grammar = Grammar(
                infrastructure=self,
                proximity_radius=proximity_radius,
                search_radius=search_radius
            )
            grammar.apply(report=report)
            if report is True:
                print("baking is done.")
            self.heuristic_paths.create(infrastructure=self)
        else:
            print("already baked.")


if __name__ == "__main__":
    infrastructure = Infrastructure()
    infrastructure.add_street(pos_1=[0, 0], pos_2=[10, 10])
    infrastructure.bake()
    print(infrastructure)
    #infrastructure.show()



from piperabm.tools.nx_query import NxGet


class Get(NxGet):
    """
    Get attributes from network elements
    """

    def get_action_queue(self, id: int):
        """
        Get agent action queue
        """
        return self.actions[id]
    
    def get_pos(self, id: int):
        """
        Get node position value
        """
        return [self.get_node_attribute(id, 'x', None), self.get_node_attribute(id, 'y', None)]
    
    def get_node_type(self, id: int) -> str:
        """
        Get node *type* value
        """
        return self.get_node_attribute(id=id, attribute='type')

    def get_edge_type(self, ids: list) -> str:
        """
        Get edge *type* value
        """
        return self.get_edge_attribute(ids=ids, attribute='type')
    
    def get_node_name(self, id: int) -> str:
        """
        Get node *name* value
        """
        return self.get_node_attribute(id=id, attribute='name')
    
    def get_edge_name(self, ids: list) -> str:
        """
        Get edge *name* value
        """
        return self.get_edge_attribute(ids=ids, attribute='name')
    
    def get_resource(self, id: int, name: str) -> float:
        """
        Get agent *resource* value
        """
        return self.get_node_attribute(id=id, attribute=name)
    
    def get_resources(self, id: int) -> dict:
        """
        Get agent resources value
        """
        result = {}
        for name in self.resource_names:
            result[name] = self.get_resource(id=id, name=name)
        return result
    
    def get_enough_resource(self, id: int, name: str) -> float:
        """
        Get agent *enough_resource* value
        """
        attribute = 'enough_' + name
        return self.get_node_attribute(id=id, attribute=attribute)

    def get_alive(self, id: str) -> bool:
        """
        Get agent *alive* value
        """
        return self.get_node_attribute(id=id, attribute='alive')
    
    def get_socioeconomic_status(self, id: str) -> bool:
        """
        Get agent *socioeconomic_status* value
        """
        return self.get_node_attribute(id=id, attribute='socioeconomic_status')
    
    def get_income(self, id: str) -> float:
        """
        Get agent *income* value
        """
        return self.get_socioeconomic_status(id=id) * self.average_income
    
    def get_current_node(self, id: str) -> int:
        """
        Get agent *current_node* value
        """
        return self.get_node_attribute(id=id, attribute='current_node')
    
    def get_home_id(self, id: str) -> int:
        """
        Get agent *home_id* value
        """
        return self.get_node_attribute(id=id, attribute='home_id')
    
    def get_balance(self, id: int) -> float:
        """
        Get agent *balance* value
        """
        return self.get_node_attribute(id=id, attribute='balance')


from piperabm.tools.fit_3d import Fit3D
from piperabm.tools.coordinate.projection import latlong_xy


class ClimateImpact(Fit3D):

    def __init__(self, coeff: float = 1):
        self.coeff = coeff
        super().__init__()
    
    def add_xy(
            self,
            x: float,
            y: float,
            temperature: float
        ) -> None:
        """
        Add new data point by x and y coordiante
        """
        self.add(x=x, y=y, z=temperature)
    
    def add_latlong(
            self,
            latitude: float,
            longitude: float,
            temperature: float,
            latitude_0: float,
            longitude_0: float
        ) -> None:
        """
        Add new data point by latitude and longitude coordiante
        """
        x, y = latlong_xy(
            latitude_0=latitude_0,
            longitude_0=longitude_0,
            latitude=latitude,
            longitude=longitude
        )
        self.add_xy(x=x, y=y, temperature=temperature)

    def temperature(self, x: float, y: float):
        """
        Interpolate temperature
        """
        return self.z(x=x, y=y)
    
    def impact(self, x: float, y: float):
        """
        Interpolate climate impact
        """
        return self.temperature(x=x, y=y) * self.coeff
    

if __name__ == "__main__":
    cliamte_impact = ClimateImpact(coeff=0.01)
    cliamte_impact.add_xy(x=10, y=0, temperature=10)
    cliamte_impact.add_xy(x=-10, y=0, temperature=-10)
    cliamte_impact.add_xy(x=0, y=10, temperature=5)
    cliamte_impact.add_xy(x=0, y=-10, temperature=0)
    print(cliamte_impact.impact(x=10, y=10))


import os
import shutil
import subprocess


class Animation:
    """
    Manages animation rendering using plt
    """

    def __init__(self, path):
        self.file_names = []
        self.path = path
        self.render_path = os.path.join(self.path, 'render')
        self._clear_folder(self.render_path)
        os.makedirs(self.render_path, exist_ok=True)

    def add_name(self, name):
        """
        Append new filename
        """
        self.file_names.append(name)

    def add_figure(self, fig):
        """
        Add new plt figure to the instance
        """
        file_name = self.new_name()
        file_path = os.path.join(self.render_path, file_name)
        fig.savefig(file_path)
        self.file_names.append(file_path)

    def new_name(self):
        """
        Generate new filename
        """
        length = len(self.file_names)
        name = f"image_{length + 1:04d}.png"
        return name

    def render(self, output_file='output', framerate=10):
        """
        Create animation
        """
        output_file = os.path.join(self.path, output_file + '.mp4')
        
        if not self.file_names:
            print("No images to render.")
            return

        # Command to combine images into a video using ffmpeg
        cmd = [
            'ffmpeg', '-y', '-framerate', str(framerate), '-i',
            os.path.join(self.render_path, 'image_%04d.png'),
            '-c:v', 'libx264', '-pix_fmt', 'yuv420p', output_file
        ]

        subprocess.run(cmd, check=True)
        print(f"Rendered video saved as {output_file}")

        # Clear the render folder after rendering
        self._clear_folder(self.render_path)

    def _clear_folder(self, folder):
        """
        Delete a folder
        """
        if os.path.exists(folder):
            shutil.rmtree(folder)


if __name__ == '__main__':

    import numpy as np
    from matplotlib import pyplot as plt


    class Model:

        def __init__(self, pos, velocity, step_size):
            self.pos = np.array(pos)
            self.velocity = np.array(velocity)
            self.step_size = step_size

        def update(self):
            self.pos += self.velocity * self.step_size

        def fig(self):
            plt.clf()
            ax = plt.gca()
            ax.set_aspect("equal")
            ax.scatter(self.pos[0], self.pos[1])
            ax.set_xlim(-10, 10)
            ax.set_ylim(-10, 10)
            return plt.gcf()
        
        def show(self):
            fig = self.fig()
            plt.show()

    
    model = Model(
        pos=[-10, -10], # meter
        velocity=[1, 1], # meter / second
        step_size=1 # second
    )

    path = os.path.dirname(os.path.realpath(__file__))
    animation = Animation(path)

    for _ in range(21):
        fig = model.fig()
        animation.add_figure(fig)
        model.update()

    animation.render()



def accessibility(resource: float, enough_resource: float) -> float:
    return min(resource / enough_resource, 1)


if __name__ == "__main__":
    resource = 5
    enough_resource = 10
    print(
        accessibility(
            resource=resource,
            enough_resource=enough_resource
        )
    )


import os
import csv
import keepdelta as kd

from piperabm.tools.json_file import JsonFile


class File:
    """
    Manages file handling methods of model
    """

    @property
    def result_directory(self):
        """
        Return result directory
        """
        if self.path is None:
            raise ValueError("Define path to continue.")
        result = os.path.join(self.path, 'result')
        if self.name != '':
            result = os.path.join(result, self.name)
        return result

    def save(self, name: str) -> None:
        """
        Save model to file
        """
        path = self.result_directory
        if not os.path.exists(path): # Create result directory if doesn't exist
            os.makedirs(path)
        data = self.serialize()
        file = JsonFile(path=path, filename=name)
        file.save(data)

    def save_initial(self) -> None:
        """
        Save initial state of model to file
        """
        self.save(name='initial')

    def save_final(self) -> None:
        """
        Save final state of model to file
        """
        self.save(name='final')

    def load(self, name: str) -> None:
        """
        Load model from file
        """
        file = JsonFile(path=self.result_directory, filename=name)
        if file.exists() is False:
            raise ValueError(f"File {name} doesn't exist in {self.result_directory}")
        data = file.load()
        self.deserialize(data)

    def load_initial(self) -> None:
        """
        Load initial state of model from file
        """
        self.load(name='initial')

    def load_final(self) -> None:
        """
        Load final state of model from file
        """
        self.load(name='final')

    def load_deltas(
            self,
            _from: int = None,
            _to: int = None
        ) -> list:
        """
        Load all detlas from file
        """
        deltas_file = JsonFile(path=self.result_directory, filename='simulation')
        deltas = deltas_file.load()
        if _from is None:
            _from = 0
        if _to is None:
            _to = len(deltas)
        deltas = deltas[_from: _to]
        return deltas

    def append_delta(self, delta) -> None:
        """
        Append the new delta to file
        """
        deltas_file = JsonFile(path=self.result_directory, filename='simulation')
        if deltas_file.exists() is False:
            deltas_file.save(data=[])
        deltas_file.append(delta)

    def apply_delta(self, delta) -> None:
        """
        Update model by applying a delta
        """
        self.deserialize(
            kd.apply(
                old=self.serialize(),
                delta=delta
            )
        )

    def apply_deltas(self, deltas: list = None) -> None:
        """
        Update model by applying all deltas
        """
        if deltas is None:
            deltas = self.load_deltas()
        for delta in deltas:
            self.apply_delta(delta)

    def push(self, steps: int = 1) -> None:
        """
        Push model forward using deltas
        """
        current = self.step
        deltas = self.load_deltas(_from=current, _to=current+steps)
        self.apply_deltas(deltas=deltas)

    def append_transactions(self, transactions) -> None:
        """
        Append the new delta to file
        """
        name = 'transactions' + '.' + 'csv'
        filepath = os.path.join(self.result_directory, name)
        headers = ['from', 'to', 'amount', 'resource', 'time']
        if not os.path.exists(filepath):
            # Create it with headers only
            with open(filepath, mode='w', newline='') as file:
                writer = csv.writer(file)
                writer.writerow(headers)
        with open(filepath, mode='a', newline='') as file:
            writer = csv.writer(file)
            for transaction in transactions:
                writer.writerow(transaction)


if __name__ == "__main__":

    from piperabm.infrastructure.samples.infrastructure_0 import model

    model.path = os.path.dirname(os.path.realpath(__file__))
    model.save_initial()
    model.load_initial()


class Update:
    """
    Update the network
    """

    def update(self, duration: float, measure: bool = False):
        """
        Update the network
        """
        # Idle resource consumption & income
        for id in self.alives:
            # Resources
            for name in self.resource_names:
                resource_value = self.get_resource(id=id, name=name)
                resource_consumption_rate = self.idle_resource_rates[name]
                new_resource_value = resource_value - (resource_consumption_rate * duration)
                self.set_resource(id=id, name=name, value=new_resource_value)
            # Income
            balance = self.get_balance(id)
            income = self.get_income(id)
            new_balance = balance + income * duration
            self.set_balance(id, value=new_balance)

        # Action update
        for id in self.alives:
            action_queue = self.actions[id]

            if action_queue.done is True:
                action_queue.reset()
                # Decide
                self.decide_destination(agent_id=id, duration=duration)

            # Execute
            action_queue.update(duration, measure=measure)


if __name__ == "__main__":

    from piperabm.model import Model

    model = Model()
    model.infrastructure.add_home(pos=[0, 0])
    model.bake()
    model.society.generate(num=1)

    print(f"deads: {len(model.society.deads)}")

    model.society.update(1000000)
    
    print(f"deads: {len(model.society.deads)}")


import csv
import os


class CsvFile:
    """
    Json file handling
    """

    def exists(self):
        """
        Check if the file already exists
        """
        return os.path.exists(self.filepath)
    


if __name__ == "__main__":
    path = os.path.dirname(os.path.realpath(__file__))
    filename = "sample"
    file = CsvFile(path, filename)

    data = []
    file.save(data)

    entry = {"a": 1}
    file.append(entry)

    data = file.load()
    print("Test: ", data == [{"a": 1}])

    file.remove()


from piperabm.tools.nx_query import NxSet


class Set(NxSet):
    """
    Set attributes to network elements
    """

    def set_pos(self, id: int, value: list) -> None:
        """
        Set node position
        """
        x = float(value[0])
        y = float(value[1])
        self.set_node_attribute(id=id, attribute='x', value=x)
        self.set_node_attribute(id=id, attribute='y', value=y)

    def set_resource(self, id: int, name: str, value: float) -> None:
        """
        Set agent *resource* value
        """
        if value <= 0:
            value = 0
            self.set_node_attribute(id=id, attribute='alive', value=False)
        self.set_node_attribute(id=id, attribute=name, value=value)

    def set_resources(self, id: int, values: dict) -> None:
        """
        Set agent resources values
        """
        for name in values:
            self.set_resource(id=id, name=name, value=values[name])

    def set_current_node(self, id: str, value: int) -> None:
        """
        Set agent *current_node* value
        """
        return self.set_node_attribute(id=id, attribute='current_node', value=value)
    
    def set_balance(self, id: str, value: float) -> None:
        """
        Set agent *balance* value
        """
        if value < 0:
            value = 0
        return self.set_node_attribute(id=id, attribute='balance', value=value)
    


"""
Default values
"""

from copy import deepcopy


""" Idle consumption """
idle_resource_rates = {
    'food': 2 / (60 * 60 * 24), # kg/day to kg/s
    'water': 2 / (60 * 60 * 24), # kg/day to kg/s
    'energy': 2 / (60 * 60 * 24), # kg/day to kg/s
}

""" Walk """
speed = 5 * ((1000) / (60 * 60)) # km/hour to m/s
transportation_resource_rates = {
    'food': 2 / (60 * 60 * 24), # kg/day to kg/s
    'water': 1 / (60 * 60 * 24), # kg/day to kg/s
    'energy': 1 / (60 * 60 * 24), # kg/day to kg/s
}

""" Resource """
average_resources = {
    'food': 20, # kg
    'water': 20, # kg
    'energy': 20, # kg
}
average_enough_resources = deepcopy(average_resources)



from piperabm.tools.average.arithmetic import arithmetic
from piperabm.tools.average.geometric import geometric


class average:

    def arithmetic(values: list, weights: list = None):
        return arithmetic(values=values, weights=weights)
    
    def geometric(values: list, weights: list = None):
        return geometric(values=values, weights=weights)


class NxGet:
    """
    General get methods for networkx
    """

    def get_node_attribute(self, id: int, attribute: str, default=None):
        """
        Get node attribute from networkx graph
        """
        return self.G.nodes[id].get(attribute, default)
    
    def get_edge_attribute(self, ids: list, attribute: str, default=None):
        """
        Get edge attribute from networkx graph
        """
        return self.G.edges[*ids].get(attribute, default)
    
    def get_node_attributes(self, id: list) -> dict:
        """
        Get all node attribute from networkx graph
        """
        return self.G.nodes[id]
    
    def get_edge_attributes(self, ids: list) -> dict:
        """
        Get all edge attribute from networkx graph
        """
        return self.G.get_edge_data(*ids)
    
    @property
    def nodes(self) -> list:
        """
        Return all nodes id
        """
        return list(self.G.nodes())
    
    @property
    def edges(self) -> list:
        """
        Return all edges ids
        """
        return list(self.G.edges())
    
    def edges_from(self, id: int) -> list:
        """
        All edges from a node
        """
        return list(self.G.edges(id))



import numpy as np


class Serialize:
    """
    Serialization methods
    """

    def serialize(self) -> dict:
        """
        Serialize
        """
        data = {}
        data['time'] = self.time
        data['step'] = self.step
        data['infrastructure'] = self.infrastructure.serialize()
        data['society'] = self.society.serialize()
        data['name'] = self.name
        data['prices'] = self.prices
        data['seed'] = self.seed
        data['state'] = self.serialize_state()
        data['type'] = self.type
        return data
    
    def deserialize(self, data: dict) -> None:
        """
        Deserialize
        """
        self.time = data['time']
        self.step = data['step']
        self.infrastructure.deserialize(data['infrastructure'])
        self.society.deserialize(data['society'])
        self.set_seed(data['seed'])
        self.deserialize_state(state=data['state'])
        self.name = data['name']
        self.prices = data['prices']

    def serialize_state(self):
        """
        Serialize numpy random generator state
        """
        state = np.random.get_state()
        return [
            state[0],
            state[1].tolist(),  # Convert numpy array to list
            state[2],
            state[3],
            state[4]
        ]
    
    def deserialize_state(self, state):
        """
        Deserialize numpy random generator state
        """
        restored_state = (
            state[0],
            np.array(state[1], dtype=np.uint32),  # Convert list back to numpy array
            state[2],
            state[3],
            state[4]
        )
        np.random.set_state(restored_state)


if __name__ == "__main__":

    import piperabm as pa

    model = pa.Model(seed=1)
    #print(model.infrastructure.new_id())
    #print(model.infrastructure.new_id())
    data = model.serialize()

    print(data)

    model_new = pa.Model()
    model_new.deserialize(data)
    #print(model_new.infrastructure.new_id())
    #print(model_new.infrastructure.new_id())
    #print(model_new.seed)


import numpy as np

from piperabm.model.serialize import Serialize
from piperabm.model.file import File
from piperabm.model.update import Update
from piperabm.model.graphics import Graphics
from piperabm.infrastructure import Infrastructure
from piperabm.society import Society


class Model(
    Serialize,
    File,
    Update,
    Graphics
):
    """
    Main class of simulation
    """

    type = "model"

    def __init__(
        self,
        name: str = '',
        prices: dict = {
                'food': 1,
                'water': 1,
                'energy': 1,
            },
        path=None,
        seed=None
    ):
        super().__init__()
        self.time = 0
        self.step = 0
        self.infrastructure = Infrastructure()
        self.infrastructure.model = self # Binding
        self.society = Society()
        self.society.model = self # Binding
        self.name = name
        self.prices = prices
        self.path = path # File saving
        self.set_seed(seed=seed)

    def set_seed(self, seed: int = None):
        """
        Set random generator seed for result reproducability
        """
        self.seed = seed
        np.random.seed(seed)

    @property
    def resource_names(self):
        """
        Return name of resources in the model
        """
        return list(self.prices.keys())

    def bake(
            self,
            save: bool = False,
            proximity_radius: float = 0,
            search_radius: float = None,
            report: bool = False
        ):
        """
        Bake model
        """
        self.infrastructure.bake(
            report=report,
            proximity_radius=proximity_radius,
            search_radius=search_radius
        )
        if save is True:
            self.save(state='infrastructure')
    

if __name__ == "__main__":

    from piperabm.model import Model

    model = Model()
    model.infrastructure.add_home(pos=[0, 0])
    model.infrastructure.add_street(pos_1=[-5, 0], pos_2=[5, 0])
    model.bake()
    print(model.infrastructure)



def arithmetic(values: list, weights: list = None):
    result = None
    if weights is None:
        #print(values)
        result = sum(values) / len(values)
    else:
        if len(values) != len(weights):
            raise ValueError("The lengths of values and weights must be the same.")
        weighted_sum = sum(v * w for v, w in zip(values, weights))
        total_weight = sum(weights)
        if total_weight == 0:
            raise ValueError("The sum of weights must not be zero.")
        result = weighted_sum / total_weight
    return result


if __name__ == "__main__":
    values = [1, 2, 3]
    print(arithmetic(values=values))

    weights = [1, 1, 1]
    print(arithmetic(values=values, weights=weights))

    weights = [100, 1, 1]
    print(arithmetic(values=values, weights=weights))


class Degradation:
    """
    Manage edge dergradation methods
    """

    def calculate_adjustment_factor(self, usage_impact: float, climate_impact: float) -> float:
        """
        Calculate adjustment factor
        """
        return 1 + (self.coeff_usage * usage_impact) + (self.coeff_weather * climate_impact)
    
    def adjustment_factor(self, ids: list) -> float:
        """
        Return edge *adjustment_factor*
        """
        return self.calculate_adjustment_factor(
            usage_impact=self.get_usage_impact(ids=ids),
            climate_impact=self.get_climate_impact(ids=ids)
        )
    
    def calculate_adjusted_length(self, length: float, adjustment_factor: float) -> float:
        """
        Calculate edge *adjusted_length*
        """
        return length * adjustment_factor
    
    def adjusted_length(self, ids: list) -> float:
        """
        Return edge *adjusted_length*
        """
        return self.calculate_adjusted_length(
            length=self.get_length(ids=ids),
            adjustment_factor=self.adjustment_factor(ids=ids)
        )
    
    def update_adjusted_length(self, ids: list):
        """
        Update *adjusted_length* value
        """
        adjusted_length = self.adjusted_length(ids=ids)
        self.set_adjusted_length(ids=ids, value=adjusted_length)

    def top_degraded_edges(self, percent: float = 0):
        """
        Filter most degradaded edges by their length percentage
        """
        if percent > 100:
            raise ValueError("enter a value between 0 and 100")
        edges_ids = self.streets
        total_length = 0
        edges_info = []
        for edge_ids in edges_ids:
            length = self.get_edge_attribute(ids=edge_ids, attribute='length')
            edge_info = {
                'ids': edge_ids,
                'degradation': self.adjustment_factor(ids=edge_ids),
                'length': length
            }
            edges_info.append(edge_info)
            total_length += length
        sorted_edges_info = sorted(edges_info, key=lambda x: x['degradation'], reverse=True)
        remaining_length = (percent / 100) * total_length
        result = []
        for edge_info in sorted_edges_info:
            remaining_length -= edge_info['length']
            if remaining_length < 0:
                break
            else:
                result.append(edge_info['ids'])
        return result


from piperabm.society.query.add import Add
from piperabm.society.query.get import Get
from piperabm.society.query.set import Set


class Query(Add, Get, Set):
    """
    Query network elements
    """

    def has_node(self, id: int) -> bool:
        """
        Check whether the network already contains the node
        """
        return self.G.has_node(id)

    def has_edge(self, ids: list) -> bool:
        """
        Check whether the network already contains the edge
        """
        return self.G.has_edge(*ids)
    
    def is_home(self, id: int) -> bool:
        return self.get_home_id(id=id) == self.get_current_node(id=id)
    
    def resources_value(self, id: int) -> float:
        """
        Monetary value of resources that an agent possesses
        """
        values = []
        for name in self.resource_names:
            value = self.get_resource(id=id, name=name) * self.prices[name]
            values.append(value)
        return sum(values)
    
    def resources_in(self, node_id, is_market: bool):
        """
        All resources available in a node
        """
        agents = self.agents_in(id=node_id)
        result = {}
        for name in self.resource_names:
            result[name] = 0
        for agent_id in agents:
            for name in self.resource_names:
                result[name] += self.get_resource(id=agent_id, name=name)
        if is_market is True:
            for name in self.resource_names:
                result[name] += self.infrastructure.get_resource(id=node_id, name=name)
        return result
        '''
        food = 0
        water = 0
        energy = 0
        for agent_id in agents:
            food += self.get_resource(id=agent_id, name='food')
            water += self.get_resource(id=agent_id, name='water')
            energy += self.get_resource(id=agent_id, name='energy')
        if is_market is True:
            food += self.infrastructure.get_resource(id=node_id, name='food')
            water += self.infrastructure.get_resource(id=node_id, name='water')
            energy += self.infrastructure.get_resource(id=node_id, name='energy')
        return food, water, energy
        '''
        
    def wealth(self, id: int) -> float:
        """
        Wealth of an agent
        """
        return self.get_balance(id) + self.resources_value(id)
    
    def ego(self, id: int, type: str = None) -> list:
        """
        Return agent nodes in the ego network
        """
        connected_nodes = list(self.G.neighbors(id))  # Get all nodes connected to node_id
        if type is not None:
            # Filter connected nodes by the edge attribute 'type' equal to 'family'
            result = []
            for neighbor in connected_nodes:
                edges = self.G.get_edge_data(id, neighbor)
                for key, attr in edges.items():
                    if attr.get('type') == type:
                        result.append(neighbor)
                        break  # If one edge meets the condition, we can stop checking further
            connected_nodes = result
        return connected_nodes

    def agents_in(self, id: int) -> list:
        """
        Return all agents in a certain node
        """
        try:
            return [n for n, attr in self.G.nodes(data=True) if attr.get('type') == 'agent' and attr.get('current_node') == id]
        except:
            return []

    def agents_from(self, home_id: int) -> list:
        """
        Return all agents from a certain home (family)
        """
        try:
            return [n for n, attr in self.G.nodes(data=True) if attr.get('type') == 'agent' and attr.get('home_id') == home_id]
        except:
            return []

    @property
    def agents(self) -> list:
        """
        Return all agent nodes
        """
        try:
            return [n for n, attr in self.G.nodes(data=True) if attr.get('type') == 'agent']
        except:
            return []
        
    @property
    def alives(self) -> list:
        """
        Return all alive agent nodes
        """
        try:
            return [n for n, attr in self.G.nodes(data=True) if attr.get('type') == 'agent' and attr.get('alive') is True]
        except:
            return []
        
    @property
    def deads(self) -> list:
        """
        Return all alive agent nodes
        """
        try:
            return [n for n, attr in self.G.nodes(data=True) if attr.get('type') == 'agent' and attr.get('alive') is False]
        except:
            return []

    @property
    def families(self) -> list:
        """
        Return all family edges
        """
        try:
            return [(u, v) for u, v, attr in self.G.edges(data=True) if attr.get('type') == 'family']
        except:
            return []
        
    @property
    def friends(self) -> list:
        """
        Return all friend edges
        """
        try:
            return [(u, v) for u, v, attr in self.G.edges(data=True) if attr.get('type') == 'friend']
        except:
            return []
        
    @property
    def neighbors(self) -> list:
        """
        Return all friend edges
        """
        try:
            return [(u, v) for u, v, attr in self.G.edges(data=True) if attr.get('type') == 'neighbor']
        except:
            return []


import numpy as np

from piperabm.tools.vector import vector as vc


def point_to_line_segment(
        point,
        line_point_1,
        line_point_2,
        perpendicular_only: bool = False,
        vector: bool = False,
        ndarray: bool = False
    ):
    result = None

    x = point[0]
    y = point[1]
    x1 = line_point_1[0]
    y1 = line_point_1[1]
    x2 = line_point_2[0]
    y2 = line_point_2[1]

    A = x - x1
    B = y - y1
    C = x2 - x1
    D = y2 - y1

    dot = A * C + B * D
    len_sq = C * C + D * D
    param = -1
    if len_sq != 0:  # in case of 0 length line
        param = dot / len_sq

    xx = None
    yy = None
    if perpendicular_only is False:
        if param < 0:
            xx = x1
            yy = y1
        elif param > 1:
            xx = x2
            yy = y2
        else:
            xx = x1 + param * C
            yy = y1 + param * D
    elif perpendicular_only is True:
        if param <= 1 and \
            param >= 0:
            xx = x1 + param * C
            yy = y1 + param * D
        
    if xx is not None and \
        yy is not None:
        dx = xx - x
        dy = yy - y
        result = [dx, dy]
    
    if result is not None:
        if vector is True:
            if ndarray is True:
                result = np.array(result)
            else:
                pass
        else:
            result = vc.magnitude(result)

    return result


if __name__ == "__main__":
    point = [-3, 4]
    line_point_1 = [0, 0]
    line_point_2 = [2, 0]
    vector = point_to_line_segment(point, line_point_1, line_point_2, perpendicular_only=False)
    print(vector)



from piperabm.society.actions.action_queue import ActionQueue
from piperabm.tools.nx_serializer import nx_serialize, nx_deserialize


class Serialize:
    """
    Serialization methods
    """

    def serialize(self) -> dict:
        """
        Serialize
        """
        data = {}
        actions_serialized = {}
        for id in self.actions:
            action_queue = self.actions[id]
            actions_serialized[id] = action_queue.serialize()
        data["actions"] = actions_serialized
        data["G"] = nx_serialize(self.G)
        data["average_income"] = self.average_income
        data["neighbor_radius"] = self.neighbor_radius
        data["max_time_outside"] = self.max_time_outside
        data["activity_cycle"] = self.activity_cycle
        data["idle_resource_rates"] = self.idle_resource_rates
        data["transportation_resource_rates"] = self.transportation_resource_rates
        data["speed"] = self.speed
        data["transportation_degradation"] = self.transportation_degradation
        data["type"] = self.type
        return data
    
    def deserialize(self, data: dict) -> None:
        """
        Deserialize
        """
        actions_serialized = data["actions"]
        self.actions = {}
        for id in actions_serialized:
            action_queue = ActionQueue(id)
            action_queue.society = self # Binding
            action_queue.deserialize(actions_serialized[id])
            self.actions[id] = action_queue
        self.G = nx_deserialize(data["G"])
        self.average_income = data["average_income"]
        self.neighbor_radius = data["neighbor_radius"]
        self.max_time_outside = data["max_time_outside"]
        self.activity_cycle = data["activity_cycle"]
        self.idle_resource_rates = data["idle_resource_rates"]
        self.transportation_resource_rates = data["transportation_resource_rates"]
        self.transportation_degradation = data["transportation_degradation"]
        self.speed = data["speed"]


if __name__ == "__main__":
    
    from piperabm.infrastructure.samples.infrastructure_0 import model

    model.set_seed(1)
    model.society.generate(num=2)
    society_serialized = model.society.serialize()

    print("society serialized:\n", society_serialized)


"""
A single home standing in the middle of nowhere
"""

import piperabm as pa


model = pa.Model()
model.set_seed(2)
model.infrastructure.coeff_usage = 1
model.infrastructure.coeff_weather = 1
model.infrastructure.add_home(pos=[0, 0], id=0, name='home')
model.infrastructure.bake()
model.set_seed(None)


if __name__ == "__main__":
    model.infrastructure.show()



"""
Simple world with one home and one market
"""

import piperabm as pa


model = pa.Model()
model.set_seed(2)
model.infrastructure.coeff_usage = 1
model.infrastructure.coeff_weather = 1
model.infrastructure.add_street(pos_1=[0, 0], pos_2=[-60, 40], name='road')
model.infrastructure.add_home(pos=[5, 0], id=1, name='home')
model.infrastructure.add_market(
    pos=[-60, 45],
    id=2,
    name='market',
    resources={'food': 100, 'water': 100, 'energy': 100}
)
model.infrastructure.bake()
model.set_seed(None)


if __name__ == "__main__":
    model.infrastructure.show()



import numpy as np


class rotate:

    @staticmethod
    def x(
        vector,
        angle,
        unit: str = 'degree',
        rotate: str = 'axis',
        ndarray: bool = True
    ):
        """
        Rotate a 3D vector around the x-axis by a given angle in degrees.
        
        :param vector: A numpy array or list representing a 2D or 3D vector.
        :param angle: The angle of rotation.
        :param unit: The unit of angle, values: 'degree', 'radian'.
        :param rotate: The way the rotation is applied, values: 'axis', 'vector'.
        :param ndarray: The output format is ndarray or list.
        :return: Rotated vector.
        """
        result = None

        vector, angle_radians, factor = preprocess(vector, angle, unit, rotate)

        # Rotation matrix around x-axis
        rotation_matrix = np.array([
            [1, 0, 0],
            [0, np.cos(angle_radians), factor * np.sin(angle_radians)],
            [0, -1 * factor * np.sin(angle_radians), np.cos(angle_radians)]
        ])
        result = np.dot(rotation_matrix, vector)
        
        if ndarray is False:
            result = [float(num) for num in result]  # Convert np.float64 to float explicitly

        return result
    

    @staticmethod
    def y(
        vector,
        angle,
        unit: str = 'degree',
        rotate: str = 'axis',
        ndarray: bool = True
    ):
        """
        Rotate a 3D vector around the y-axis by a given angle in degrees.
        
        :param vector: A numpy array or list representing a 2D or 3D vector.
        :param angle: The angle of rotation.
        :param unit: The unit of angle, values: 'degree', 'radian'.
        :param rotate: The way the rotation is applied, values: 'axis', 'vector'.
        :param ndarray: The output format is ndarray or list.
        :return: Rotated vector.
        """
        result = None

        vector, angle_radians, factor = preprocess(vector, angle, unit, rotate)

        # Rotation matrix around y-axis
        rotation_matrix = np.array([
            [np.cos(angle_radians), 0, -1 * factor * np.sin(angle_radians)],
            [0, 1, 0],
            [factor * np.sin(angle_radians), 0, np.cos(angle_radians)]
        ])
        result = np.dot(rotation_matrix, vector)
        
        if ndarray is False:
            result = [float(num) for num in result]  # Convert np.float64 to float explicitly
            
        return result


    @staticmethod
    def z(
        vector,
        angle,
        unit: str = 'degree',
        rotate: str = 'axis',
        ndarray: bool = True
    ):
        """
        Rotate a 3D vector around the z-axis by a given angle in degrees.
        
        :param vector: A numpy array or list representing a 2D or 3D vector.
        :param angle: The angle of rotation.
        :param unit: The unit of angle, values: 'degree', 'radian'.
        :param rotate: The way the rotation is applied, values: 'axis', 'vector'.
        :param ndarray: The output format is ndarray or list.
        :return: Rotated vector.
        """
        result = None

        vector, angle_radians, factor = preprocess(vector, angle, unit, rotate)

        # Rotation matrix around z-axis
        rotation_matrix = np.array([
            [np.cos(angle_radians), factor * np.sin(angle_radians), 0],
            [-1 * factor * np.sin(angle_radians), np.cos(angle_radians), 0],
            [0, 0, 1]
        ])
        result = np.dot(rotation_matrix, vector)
        
        if ndarray is False:
            result = [float(num) for num in result]  # Convert np.float64 to float explicitly
            
        return result


def preprocess(vector, angle, unit='degree', rotate='axis'):
    
    vector = np.array(vector)
    
    while len(vector) < 3:
        vector = np.append(vector, 0)

    if rotate == 'axis':
        factor = 1
    elif rotate == 'vector':
        factor = -1

    if unit == 'radian':
        angle_radians = angle
    elif unit == 'degree':
        angle_radians = np.radians(angle)

    return vector, angle_radians, factor


if __name__ == "__main__":
    vector = np.array([0, 1, 0])
    angle = 45
    rotated_vector = rotate.x(vector, angle, unit='degree', rotate='axis')

    print(f'Original Vector: {vector}')
    print(f'Rotated Vector: {rotated_vector}')


"""
10*10 grid of points (not equally spaced) having equal latitude and longitude difference on sphere are projected
"""

import numpy as np
import matplotlib.pyplot as plt

from piperabm.tools.coordinate.projection.latlong_xy import latlong_xy


latitude_0 = 70
longitude_0 = -150
DELTA = 5 # degrees
num = 10

# Calculate latitude and longitude values
latitudes_linspace = list(np.linspace(latitude_0 - DELTA/2, latitude_0 + DELTA/2, num))
longitudes_linspace = list(np.linspace(longitude_0 - DELTA/2, longitude_0 + DELTA/2, num))
latitudes_mesh, longitudes_mesh = np.meshgrid(latitudes_linspace, longitudes_linspace)
latitudes = latitudes_mesh.flatten()
longitudes = longitudes_mesh.flatten()

# Calculate x and y values
xs = []
ys = []
for i in range(num ** 2):
    x, y = latlong_xy(latitude_0, longitude_0, latitude=latitudes[i], longitude=longitudes[i])
    xs.append(x)
    ys.append(y)

# Draw
fig, (ax1, ax2) = plt.subplots(1, 2)
fig.suptitle('latlong_to_xy')

ax1.scatter(longitudes, latitudes)
ax1.set_title('before')
ax1.set_xlabel('longitude (degree)')
ax1.set_ylabel('latitude (degree)')

ax2.scatter(xs, ys)
ax2.set_title('after')
ax2.set_xlabel('x (meters)')
ax2.set_ylabel('y (meters)')

# Display the plots
plt.tight_layout()
plt.show()



import numpy as np
from copy import deepcopy

from piperabm.tools.coordinate import distance as ds


class Add:
    """
    Add new network elements
    """

    def check_id(self, id):
        """
        Check whether id already exists
        """
        if id is None:
            id = self.new_id()
        else:
            if id in self.nodes:
                id = self.new_id()
                print("id already exists. replaced with new id.")
        return id

    def new_id(self):
        """
        Generate new unique random id
        """
        return int(np.random.randint(low=0, high=np.iinfo(np.int64).max, dtype=np.int64))

    def add_junction(
        self,
        pos: list,
        id: int = None,
        name: str = '',
        report: bool = False
    ):
        """
        Add junction node
        """
        type = 'junction'
        id = self.check_id(id)
        self.G.add_node(
            id,
            name=name,
            type=type,
            x=float(pos[0]),
            y=float(pos[1])
        )
        self.baked_streets = False
        self.baked_neighborhood = False
        if report is True:
            print(f">>> {type} node at position {pos} added.")
        return id
    
    def add_home(
        self,
        pos: list,
        id: int = None,
        name: str = '',
        report: bool = False
    ):
        """
        Add home node
        """
        type = 'home'
        id = self.check_id(id)
        self.G.add_node(
            id,
            name=name,
            type=type,
            x=float(pos[0]),
            y=float(pos[1])
        )
        self.baked_streets = False
        self.baked_neighborhood = False
        if report is True:
            print(f">>> {type} node at position {pos} added.")
        return id
    
    def add_market(
        self,
        pos: list,
        resources: dict = {
            'food': 10,
            'water': 10,
            'energy': 10,
        },
        enough_resources: dict = None,
        id: int = None,
        name: str = '',
        report: bool = False,
    ):
        """
        Add market node
        """
        type = 'market'
        id = self.check_id(id)
        resource_kwargs = {}
        if enough_resources is None:
            enough_resources = {}
            for resource_name in self.resource_names:
                enough_resources[resource_name] = None
        for resource_name in self.resource_names:
            if enough_resources[resource_name] is None:
                enough_resources[resource_name] = deepcopy(resources[resource_name])
            resource_kwargs[resource_name] = resources[resource_name]
            resource_kwargs['enough_'+resource_name] = enough_resources[resource_name]
        
        self.G.add_node(
            id,
            name=name,
            type=type,
            x=float(pos[0]),
            y=float(pos[1]),
            balance=0,
            **resource_kwargs
        )

        self.baked_streets = False
        self.baked_neighborhood = False
        if report is True:
            print(f">>> {type} node at position {pos} added.")
        return id

    def add_street(
        self,
        pos_1: list,
        pos_2: list,
        name: str = '',
        usage_impact: float = 0,
        climate_impact: float = 0,
        report: bool = False
    ):
        """
        Add street edge
        """
        type = 'street'
        id_1 = self.add_junction(pos=pos_1)
        id_2 = self.add_junction(pos=pos_2)
        length = ds.point_to_point(pos_1, pos_2)
        adjustment_factor = self.calculate_adjustment_factor(
            usage_impact=usage_impact,
            climate_impact=climate_impact
        )
        adjusted_length = self.calculate_adjusted_length(
            length=length,
            adjustment_factor=adjustment_factor
        )
        self.G.add_edge(
            id_1,
            id_2,
            name=name,
            length=length,
            adjusted_length=adjusted_length,
            usage_impact=usage_impact,
            climate_impact=climate_impact,
            type=type,
        )
        self.baked_streets = False
        self.baked_neighborhood = False
        if report is True:
            print(f">>> {type} edge at positions {pos_1}-{pos_2} added.")
        return id
    
    def add_neighborhood_access(
        self,
        id_1: list,
        id_2: list,
        name: str = '',
        usage_impact: float = 0,
        climate_impact: float = 0,
        report: bool = False
    ):
        """
        Add neighborhood access edge
        """
        type = 'neighborhood_access'
        length = ds.point_to_point(self.get_pos(id_1), self.get_pos(id_2))
        adjustment_factor = self.calculate_adjustment_factor(
            usage_impact=usage_impact,
            climate_impact=climate_impact
        )
        adjusted_length = self.calculate_adjusted_length(
            length=length,
            adjustment_factor=adjustment_factor
        )
        self.G.add_edge(
            id_1,
            id_2,
            name=name,
            length=length,
            adjusted_length=adjusted_length,
            usage_impact=usage_impact,
            climate_impact=climate_impact,
            type=type
        )
        #self.baked_streets = False
        self.baked_neighborhood = False
        if report is True:
            print(f">>> {type} edge at positions {self.get_pos(id_1)} - {self.get_pos(id_2)} added.")
        return id


import os

import piperabm as pa


path = os.path.dirname(os.path.realpath(__file__))
measurement = pa.Measurement(path=path)
measurement.load()


if __name__ == "__main__":
    _from = None
    _to = None
    agents='all'
    resources='all'
    measurement.accessibility.show(agents, resources, _from, _to)
    measurement.travel_distance.show(_from, _to)



from piperabm.tools.coordinate.distance.point_to_point import point_to_point
from piperabm.tools.coordinate.distance.point_to_line import point_to_line
from piperabm.tools.coordinate.distance.point_to_line_segment import point_to_line_segment


class distance:

    @staticmethod
    def point_to_point(point_1, point_2, vector=False, ndarray=False):
        return point_to_point(point_1, point_2, vector, ndarray)

    @staticmethod
    def point_to_line(point, line, segment=False, perpendicular_only=False, vector=False, ndarray=False):
        result = None

        line_point_1 = line[0]
        line_point_2 = line[1]

        if segment is False:
            result = point_to_line(point, line_point_1, line_point_2, vector, ndarray)
        elif segment is True:
            result = point_to_line_segment(point, line_point_1, line_point_2, perpendicular_only, vector, ndarray)

        return result



"""
Solves a single resource for multiple players at fixed price
"""

from scipy.optimize import minimize
import warnings

from piperabm.economy.accessibility import accessibility


warnings.filterwarnings('ignore', message="Values in x were outside bounds during a minimize step, clipping to bounds")


class NashBargaining:

    def transactions(players: list, price: float) -> dict:
        """
        Calculate the list of transfers for proper resource allocation
        """
        num_players = len(players)

        if num_players == 0:
            return {
                'resource': [],
                'money': []
            }
        
        def objective(resource_transafers):
            """
            Optimization objective function
            """
            accessibilities = [
                accessibility(
                    players[i]['resource'] + resource_transafers[i],
                    players[i]['enough_resource']
                )
                for i in range(num_players)
            ]
            result = 1
            for a in accessibilities:
                result *= a
            return -result
        
        initial_guess = []
        bounds = []
        for i in range(num_players):
            need = max(0, players[i]['enough_resource'] - players[i]['resource'])
            possible = players[i]['balance'] / price
            upper_bound = min(need, possible)
            lower_bound = -players[i]['resource']
            bound = (lower_bound, upper_bound)
            bounds.append(bound)
            initial_guess.append((upper_bound + lower_bound) / 2)
        constraints = [
            {'type': 'eq', 'fun': lambda resource_transafers: sum(resource_transafers)},
        ]
        result = minimize(
            objective,
            initial_guess,
            bounds=bounds,
            constraints=constraints,
        )
        resource_transfers = [float(resource_transfer) for resource_transfer in result.x]
        money_transfers = [float(-resource_transfer * price) for resource_transfer in resource_transfers]
        return {
            'resource': resource_transfers,
            'money': money_transfers
        }
    
    def apply(players: list, transactions: dict) -> list:
        """
        Apply transfers to players
        """
        num_players = len(players)
        for i in range(num_players):
            players[i]['balance'] = players[i]['balance'] + transactions['money'][i]
            players[i]['resource'] = players[i]['resource'] + transactions['resource'][i]
        return players


if __name__ == "__main__":
    price = 10
    player_1 = {
        'id': 1,
        'type': 'agent',
        'resource': 9,
        'enough_resource': 10,
        'balance': 100,
    }
    player_2 = {
        'id': 2,
        'type': 'agent',
        'resource': 6,
        'enough_resource': 10,
        'balance': 100,
    }
    player_3 = {
        'id': 3,
        'type': 'agent',
        'resource': 3,
        'enough_resource': 10,
        'balance': 10,
    }
    players = [player_1, player_2, player_3]

    # Initial
    print("\n" + ">>> " + "Initial: ")
    for player in players:
        print(player)

    # Solve
    transactions = NashBargaining.transactions(players, price)
    players = NashBargaining.apply(players, transactions)

    # Final
    print("\n" + ">>> " + "Final: ")
    for player in players:
        print(player)



from copy import deepcopy

from piperabm.tools.coordinate import distance as ds


class Rule1:
    """
    Condition for node to edge proximity
    """

    name = "rule 1"

    def __init__(self, infrastructure, proximity_radius: float = 1):
        self.infrastructure = infrastructure
        self.proximity_radius = proximity_radius

    def check(self, node_id, edge_ids):
        result = False
        if node_id not in edge_ids:
            distance = ds.point_to_line(
                point=self.infrastructure.get_pos(node_id),
                line=[self.infrastructure.get_pos(edge_ids[0]), self.infrastructure.get_pos(edge_ids[1])],
                segment=True,
                perpendicular_only=True
            )
            if distance is not None and \
            distance < self.proximity_radius:
                result = True
                '''
                distance_1 = ds.point_to_point(
                    point_1=self.infrastructure.get_pos(node_id),
                    point_2=self.infrastructure.get_pos(edge_ids[0])
                )
                distance_2 = ds.point_to_point(
                    point_1=self.infrastructure.get_pos(node_id),
                    point_2=self.infrastructure.get_pos(edge_ids[1])
                )
                if distance_1 > self.proximity_radius and \
                distance_2 > self.proximity_radius:
                    result = True
                '''
        return result
    
    def apply(self, node_id, edge_ids, report=False):
        data = self.infrastructure.get_edge_attributes(ids=edge_ids)
        #edge_type = self.infrastructure.edge_type(ids=edge_ids)
        #degradation = self.infrastructure.edge_degradation(ids=edge_ids)
        data_1 = deepcopy(data)
        data_1['length'] = ds.point_to_point(
            self.infrastructure.get_pos(node_id),
            self.infrastructure.get_pos(edge_ids[0])
        )
        adjustment_factor = self.infrastructure.calculate_adjustment_factor(
            usage_impact=data_1['usage_impact'],
            climate_impact=data_1['climate_impact']
        )
        data_1['adjusted_length'] = self.infrastructure.calculate_adjusted_length(
            length=data_1['length'],
            adjustment_factor=adjustment_factor
        )
        self.infrastructure.G.add_edge(
            node_id,
            edge_ids[0],
            **data_1
        )
        data_2 = deepcopy(data)
        data_2['length'] = ds.point_to_point(
            self.infrastructure.get_pos(node_id),
            self.infrastructure.get_pos(edge_ids[1])
        )
        adjustment_factor = self.infrastructure.calculate_adjustment_factor(
            usage_impact=data_1['usage_impact'],
            climate_impact=data_1['climate_impact']
        )
        data_2['adjusted_length'] = self.infrastructure.calculate_adjusted_length(
            length=data_2['length'],
            adjustment_factor=adjustment_factor
        )
        self.infrastructure.G.add_edge(
            node_id,
            edge_ids[1],
            **data_2
        )
        if report is True:
            print(f">>> {data_1['type']} edge at positions {self.infrastructure.get_pos(node_id)} - {self.infrastructure.get_pos(edge_ids[0])} added.")
            print(f">>> {data_2['type']} edge at positions {self.infrastructure.get_pos(node_id)} - {self.infrastructure.get_pos(edge_ids[1])} added.")
        self.infrastructure.remove_edge(ids=edge_ids, report=report)

    def find(self, report=False):
        anything_happened = False
        nodes = self.infrastructure.junctions
        edges = self.infrastructure.edges
        for node_id in nodes:
            for edge_ids in edges:
                if self.check(node_id, edge_ids) is True:
                    if report is True:
                        print(f"# {self.name}:")
                    self.apply(node_id, edge_ids, report)
                    # Inform an activity
                    anything_happened = True
                # Inform an activity     
                if anything_happened is True:
                    break
            # Inform an activity     
            if anything_happened is True:
                break
        return anything_happened


if __name__ == "__main__":

    from piperabm.infrastructure import Infrastructure

    infrastructure = Infrastructure()
    infrastructure.add_street(pos_1=[0, 0], pos_2=[10, 0])
    infrastructure.add_junction(pos=[5, 0.5])

    rule = Rule1(infrastructure, proximity_radius=1)
    rule.find(report=True)
    print(infrastructure)



import os

import piperabm as pa
from piperabm.society.samples.society_1 import model


# Setup
path = os.path.dirname(os.path.realpath(__file__))
model.path = path
model.society.max_time_outside = 300
model.society.activity_cycle = 1000
model.society.average_income = 10
model.infrastructure.coeff_usage = 0.3

# Run
print(">>> Running...")
model.run(n=1000, save=True, resume=False, report=True, step_size=10)

# Measure
print(">>> Measuring...")
measurement = pa.Measurement(path=path)
measurement.measure(resume=False, report=True)


import numpy as np
from copy import deepcopy

from piperabm.tools.coordinate import distance as ds
from piperabm.tools.print.serialized import Print


class Track(Print):
    """
    Move action segments
    """

    type = "track"

    def __init__(
        self,
        action = None,
        id_start: int = None,
        id_end: int = None,
    ):
        super().__init__()
        self.action = action # Binding
        self.id_start = id_start
        self.id_end = id_end

    def preprocess(self, kwargs: dict = None):
        """
        Calculate the neccessary pieces using the kwargs
        """
        if kwargs is None:
            self.done = False
            self.elapsed = 0
            self.remaining = self.adjusted_length() / self.speed
            self.total_duration = deepcopy(self.remaining)
            pos_start = self.pos_start
            pos_end = self.pos_end
            if pos_start != pos_end:
                unit_vector = list(ds.point_to_point(self.pos_start, self.pos_end, vector=True, ndarray=True) / self.length)
                self.unit_vector = [float(i) for i in unit_vector]
                self.adjustment_factor = self.adjusted_length() / self.length
            else:
                self.unit_vector = None
                self.adjustment_factor = None
        else: # Equivalent of deserialization
            self.done = kwargs["done"]
            self.elapsed = kwargs["elapsed"]
            self.remaining = kwargs["remaining"]
            self.total_duration = kwargs["total_duration"]
            if kwargs["unit_vector"] is None:
                self.unit_vector = None
                self.adjustment_factor = None
            else:
                self.unit_vector = kwargs["unit_vector"]
                self.adjustment_factor = kwargs["adjustment_factor"]

    @property
    def action_queue(self):
        """
        Alias to access action queue
        """
        return self.action.action_queue
    
    @property
    def society(self):
        """
        Alias to access society
        """
        return self.action_queue.society
    
    @property
    def model(self):
        """
        Alias to access model
        """
        return self.society.model
    
    @property
    def infrastructure(self):
        """
        Alias to access infrastructure
        """
        return self.model.infrastructure

    @property
    def agent_id(self) -> int:
        """
        Alias to access agent id
        """
        return self.action_queue.agent_id

    @property
    def edge_ids(self) -> list:
        """
        Return corresponding edge ids
        """
        return [self.id_start, self.id_end]

    @property
    def pos_start(self) -> list:
        """
        Starting node position
        """
        return self.infrastructure.get_pos(id=self.id_start)

    @property
    def pos_end(self) -> list:
        """
        Ending node position
        """
        return self.infrastructure.get_pos(id=self.id_end)
    
    def pos(self, new_val: list = None) -> list:
        """
        Return current agent position
        """
        if new_val is None:
            return self.society.get_pos(id=self.agent_id)
        else:
            self.society.set_pos(id=self.agent_id, value=new_val)
    
    @property
    def speed(self) -> float:
        """
        Alias to access agent transportation speed
        """
        return self.society.speed

    @property
    def length(self) -> float:
        """
        Alias to access edge length
        """
        return self.infrastructure.get_length(ids=self.edge_ids)

    def adjusted_length(self) -> float:
        """
        Alias to access edge adjusted length
        """
        return self.infrastructure.get_adjusted_length(ids=self.edge_ids)
    
    def set_usage_impact(self, value: float) -> None:
        """
        Alias to set usage_impact value
        """
        self.infrastructure.set_usage_impact(ids=self.edge_ids, value=value)

    def update(self, duration: float, measure: bool = False):
        """
        Update track
        """
        if duration <= self.remaining and \
            self.unit_vector is not None:
            excess_duration = 0
            self.elapsed += duration
            self.remaining -= duration
            delta_length_adjusted = self.speed * duration
            delta_length = delta_length_adjusted / self.adjustment_factor
            new_pos = list(np.array(self.pos()) + np.array(self.unit_vector) * delta_length)
            self.pos(new_val=new_pos)
            if measure is True:
                delta_length
        else:
            excess_duration = duration - self.remaining
            self.elapsed = self.total_duration
            self.remaining = 0
            pos_old = deepcopy(self.pos())
            self.pos(new_val=self.pos_end)
            self.done = True
            # Update usage
            usage_impact = self.infrastructure.get_usage_impact(ids=self.edge_ids)
            delta = self.action.usage
            new_usage_impact = usage_impact + delta
            self.set_usage_impact(value=new_usage_impact)
            # Update adjusted_length
            #self.infrastructure.update_adjusted_length(ids=self.edge_ids)
            if measure is True:
                pos_new = deepcopy(self.pos())
                delta_length = ds.point_to_point(pos_old, pos_new)
        # Update fuel consumption
        delta_t = duration - excess_duration
        fuel_food = self.society.transportation_resource_rates["food"] * delta_t
        fuel_water = self.society.transportation_resource_rates["water"] * delta_t
        fuel_energy = self.society.transportation_resource_rates["energy"] * delta_t
        new_food = self.society.get_resource(name="food", id=self.agent_id) - fuel_food
        new_water = self.society.get_resource(name="water", id=self.agent_id) - fuel_water
        new_energy = self.society.get_resource(name="energy", id=self.agent_id) - fuel_energy
        self.society.set_resource(name="food", id=self.agent_id, value=new_food)
        self.society.set_resource(name="water", id=self.agent_id, value=new_water)
        self.society.set_resource(name="energy", id=self.agent_id, value=new_energy)
        return excess_duration
    
    def reverse(self):
        """
        Create a reversed track
        """
        reversed_track = Track(
            action=self.action,
            id_start=self.id_end,
            id_end=self.id_start
        )
        reversed_track.preprocess()
        return reversed_track

    def serialize(self):
        """
        Serialize
        """
        dictionary = {}
        dictionary["id_start"] = self.id_start
        dictionary["id_end"] = self.id_end
        dictionary["elapsed"] = self.elapsed
        dictionary["remaining"] = self.remaining
        dictionary["total_duration"] = self.total_duration
        if self.unit_vector is not None:
            dictionary["unit_vector"] = self.unit_vector
        else:
            dictionary["unit_vector"] = None
        dictionary["adjustment_factor"] = self.adjustment_factor
        dictionary["done"] = self.done
        dictionary["type"] = self.type
        return dictionary


import numpy as np


def gini_coefficient(values) -> float:
    """
    Compute Gini coefficient of array of values
    """
    values = np.array(values)
    values = np.sort(values)
    n = len(values)
    index = np.arange(1, n + 1)
    return float((2 * np.sum(index * values) / np.sum(values) - (n + 1)) / n)
    
    '''
    diffsum = 0
    for i, xi in enumerate(x[:-1], 1):
        diffsum += np.sum(np.abs(xi - x[i:]))
    return float(diffsum / (len(x)**2 * np.mean(x)))
    '''


if __name__ == "__main__":
    incomes = [100, 300, 500, 700, 900, 300, 500, 700, 500]
    gini_index = gini_coefficient(incomes)
    print(gini_index)



from piperabm.economy.trade.trade_solver import MultiResourceTrade as mrt


prices = {
    'food': 10,
    'water': 10,
    'energy': 10
}
player_1 = {
    'id': 1,
    'type': 'market',
    'resources': {
        'food': 100,
        'water': 100,
        'energy': 100
    },
    'enough_resources': {
        'food': 100,
        'water': 100,
        'energy': 100
    },
    'balance': 0,
}
player_2 = {
    'id': 2,
    'type': 'agent',
    'resources': {
        'food': 10,
        'water': 3,
        'energy': 10
    },
    'enough_resources': {
        'food': 10,
        'water': 10,
        'energy': 10
    },
    'balance': 100,
}
player_3 = {
    'id': 3,
    'type': 'agent',
    'resources': {
        'food': 1,
        'water': 2,
        'energy': 3
    },
    'enough_resources': {
        'food': 10,
        'water': 10,
        'energy': 10
    },
    'balance': 10,
}
players = [player_1, player_2, player_3]

players = mrt.solve(players=players, prices=prices)


if __name__ == "__main__":
    for player in players:
        print(f"resources: {player['resources']}, balance: {player['balance']}")



import os
from copy import deepcopy
import matplotlib.pyplot as plt

from piperabm.tools.average import average as avg


class TravelDistance:
    """
    Manage travel distance measurement
    """

    type = 'travel distance'

    def __init__(self, measurement):
        self.measurement = measurement
        self.values = []

    def add(self, value: float) -> None:
        """
        Add new travel distance value
        """
        self.values.append(value)

    def filter(self, _from=None, _to=None):
        """
        Filter values in a specific range
        """
        if _from is None:
            _from = 0
        if _to is None:
            _to = len(self.values)
        return self.values[_from: _to]
    
    def __call__(self, _from=None, _to=None):
        return self.filter(_from=_from, _to=_to)
    
    def average(self, _from=None, _to=None) -> float:
        """
        Calculate total average
        """
        values = self.__call__(_from=_from, _to=_to)
        weights = self.measurement.delta_times(_from=_from, _to=_to)
        result = avg.arithmetic(values=values, weights=weights)
        if isinstance(result, complex):
            result = float(result.real)
        return result
    
    def create_plot(self, _from=None, _to=None, info=None):
        """
        Create plot
        """
        fig, ax = plt.subplots()
        title = "Travel Distance"
        ylabel = deepcopy(title)
        if info is not None:
            title += info
        ax.set_title(title)
        xs = self.measurement.filter_times(_from=_from, _to=_to)
        yx = self.filter(_from=_from, _to=_to)
        ax.plot(xs, yx, color='blue')
        ax.set_xlabel("Time")
        ax.set_ylabel(ylabel)
        return fig
    
    def show(self, _from=None, _to=None, info=None):
        """
        Draw plot
        """
        fig = self.create_plot(
            _from=_from,
            _to=_to,
            info=info
        )
        plt.show()

    def save(self, _from=None, _to=None, info=None):
        """
        Save plot
        """
        fig = self.create_plot(
            _from=_from,
            _to=_to,
            info=info
        )
        path = self.measurement.result_directory
        filepath = os.path.join(path, self.type)
        fig.savefig(filepath)

    def serialize(self) -> dict:
        """
        Serialize
        """
        return {
            'values': self.values,
            'type': self.type
        }
    
    def deserialize(self, data: dict) -> None:
        """
        Deserialize
        """
        self.values = data['values']


if __name__ == "__main__":
    
    from piperabm.model.measurement import Measurement

    measure = Measurement()
    hour = 3600
    measure.add_time(0 * hour) # Base

    # 1
    measure.add_time(value=1*hour)
    measure.travel_distance.add(value=1.1)
    # 2
    measure.add_time(value=2*hour)
    measure.travel_distance.add(value=0.9)
    # 3
    measure.add_time(value=3*hour)
    measure.travel_distance.add(value=0.3)
    # 4
    measure.add_time(value=4*hour)
    measure.travel_distance.add(value=0.46)
    # 5
    measure.add_time(value=5*hour)
    measure.travel_distance.add(value=0.2)

    _from = None
    _to = None
    print("travel distances: ", measure.travel_distance(_from=_from, _to=_to))
    measure.travel_distance.show(_from=_from, _to=_to)


import networkx as nx
import matplotlib.pyplot as plt

from piperabm.society.graphics.style import society_style


class Graphics:
    """
    Handle graphics
    """

    def fig(
            self,
            relationships: bool = False,
            clf: bool = False
        ):
        """
        Add society elements to plt fig ax
        """
        if clf is True:
            plt.clf()
        ax = plt.gca()
        ax.set_aspect("equal")

        # Nodes
        pos_dict = {}
        node_color_list = []
        node_size_list = []
        node_label_dict = {}
        nodes = self.nodes
        for node_id in nodes:
            # Position
            pos_dict[node_id] = self.get_pos(id=node_id)
            # Color
            colors = society_style["node"]["agent"]["color"]
            if self.get_alive(id=node_id) is True:
                color = colors['alive']
            else:
                color = colors['dead']
            node_color_list.append(color)
            # Size
            size = society_style["node"]["agent"]["size"]
            node_size_list.append(size)
            # Label
            node_label_dict[node_id] = self.get_node_name(id=node_id)

        # Edges
        edge_color_list = []
        edges = []
        if relationships is not False:
            for u, v, key, data in self.G.edges(keys=True, data=True):
                type = data['type']
                if (isinstance(relationships, list) and type in relationships) or \
                (relationships is True):
                    # Add edge to list
                    edges.append([u, v, key])
                    # Color
                    color = society_style['edge'][type]['color']
                    edge_color_list.append(color)

        # Draw
        nx.draw_networkx(
            self.G,
            nodelist=nodes,
            pos=pos_dict,
            node_color=node_color_list,
            node_size=node_size_list,
            labels=node_label_dict,
            font_size=society_style['font'],
            edgelist=edges,
            edge_color=edge_color_list,
            ax=ax
        )
        return plt.gcf()

    def show(self, relationships=True):
        """
        Show society elements
        """
        fig = self.fig(relationships=relationships)
        plt.show()


if __name__ == "__main__":

    from piperabm.society.samples.society_2 import model

    model.society.show(relationships=["family"])


"""
Grid world without access to market
"""

from copy import deepcopy

from piperabm.infrastructure.samples.infrastructure_3 import model as model_infrastructure


model = deepcopy(model_infrastructure)
model.set_seed(2)
model.society.generate(num=10)
model.set_seed(None)


if __name__ == "__main__":
    print(model.society)
    #model.show()


from piperabm.infrastructure.grammar.rules import *


class Grammar:
    """
    Manage and apply gramamr rules
    """

    def __init__(
            self,
            infrastructure,
            proximity_radius: float = 1,
            search_radius: float = None,
        ):
        self.infrastructure = infrastructure
        if search_radius is not None and \
        search_radius < proximity_radius:
            print("search radius should be bigger than proximity radius")
            raise ValueError
        self.search_radius = search_radius
        self.proximity_radius = proximity_radius

    def apply(self, report=False):
        """
        Apply all grammar rules
        """
        # Baking streets
        if self.infrastructure.baked_streets is False:
            self.apply_street_grammar(report=report)
            self.infrastructure.baked_streets = True

        # Baking neighborhood
        if self.infrastructure.baked_neighborhood is False:
            self.apply_neighborhood_grammar(report=report)
            self.infrastructure.baked_neighborhood = True

    def apply_street_grammar(self, report=False):
        """
        Apply all street grammar rules
            if a rule is not yielding any changes, it is ok to go the next rule.
            if not, all grammars rules start over.
            if no next rule is available, the program is over.
        """

        rules = [
            Rule0(self.infrastructure, self.proximity_radius),
            Rule1(self.infrastructure, self.proximity_radius),
            Rule2(self.infrastructure, self.proximity_radius),
        ]

        i = 0
        while True:
            rule = rules[i]
            anything_happened = rule.find(report=report)
            if anything_happened is True:
                i = 0  # reset the loop
            else:
                i += 1  # move to the next grammar
            if i == len(rules): # Done
                break

    def apply_neighborhood_grammar(self, report=False):
        """
        Apply all neighborhood grammar rules
            if a rule is not yielding any changes, it is ok to go the next rule.
            if not, all grammars rules start over.
            if no next rule is available, the program is over.
        """

        rules = [
            Rule3(self.infrastructure, proximity_radius=self.proximity_radius, search_radius=self.search_radius),
        ]

        i = 0
        while True:
            rule = rules[i]
            anything_happened = rule.find(report=report)
            if anything_happened is True:
                i = 0  # reset the loop
            else:
                i += 1  # move to the next grammar
            if i == len(rules): # Done
                break


if __name__ == "__main__":

    from piperabm.infrastructure import Infrastructure

    infrastructure = Infrastructure()
    infrastructure.add_street(pos_1=[-10, 0], pos_2=[10, 0])
    infrastructure.add_street(pos_1=[0, -10], pos_2=[0, 10])
    infrastructure.add_home(pos=[5, 0.5])
    grammar = Grammar(infrastructure)
    grammar.apply(report=True)
    print(infrastructure.baked)



from piperabm.tools.nx_serializer import nx_serialize, nx_deserialize


class Serialize:
    """
    Serialization methods
    """

    def serialize(self):
        """
        Serialize
        """
        data = {}
        data['G'] = nx_serialize(self.G)
        data['coeff_usage'] = self.coeff_usage
        data['coeff_weather'] = self.coeff_weather
        data['baked_streets'] = self.baked_streets
        data['baked_neighborhood'] = self.baked_neighborhood
        data['heuristic_paths'] = self.heuristic_paths.serialize()
        data['type'] = self.type
        return data
    
    def deserialize(self, data):
        """
        Deserialize
        """
        self.G = nx_deserialize(data['G'])
        self.coeff_usage = data['coeff_usage']
        self.coeff_weather = data['coeff_weather']
        self.baked_streets = data['baked_streets']
        self.baked_neighborhood = data['baked_neighborhood']
        self.heuristic_paths.deserialize(data['heuristic_paths'])


if __name__ == "__main__":

    from piperabm.infrastructure.samples.infrastructure_0 import model

    print(model.infrastructure.serialize())


from copy import deepcopy
import keepdelta as kd

from piperabm.model.trade import Trade
from piperabm.tools.json_file import JsonFile


class Update(Trade):
    """
    Manage running simulation
    """

    def run(
            self,
            n: int = None,
            step_size: float = 3600,
            save: bool = False,
            save_transactions: bool = False,
            resume: bool = False,
            report: bool = False
        ):
        """
        Run model for multiple steps
        """
        # Remove previous save file if exists
        if self.path is not None:
            path = self.result_directory
            if resume is False:
                # Remove deltas
                simulation_file = JsonFile(path, 'simulation')
                simulation_file.remove()
                # Remove final state
                final_file = JsonFile(path, 'final')
                final_file.remove()
                # Load initial state if exists
                initial_file = JsonFile(path, 'initial')
                if initial_file.exists():
                    self.load_initial()
                else:
                    if save is True:
                        self.save_initial()
            else:
                # Load final state if exists
                final_file = JsonFile(path, 'final')
                if final_file.exists() is True:
                    self.load_final()
                else:
                    # Load initial state if final state doesn't exists
                    initial_file = JsonFile(path, 'initial')
                    if initial_file.exists() is True:
                        self.load_initial()
                        # Apply deltas if exists
                        simulation_file = JsonFile(path, 'simulation')
                        if simulation_file.exists() is True:
                            self.apply_deltas()

        # Run until all agents die
        if n is None:
            while True:
                self.update(
                    duration=step_size,
                    save=save,
                    save_transactions=save_transactions
                )
                if len(self.society.alive_agents) == 0:
                    break

        # Run for certain steps
        else:
            for i in range(n):
                if report is True:
                    print(f"Progress: {(i + 1) / n * 100:.1f}% complete")
                self.update(
                    duration=step_size,
                    save=save,
                    save_transactions=save_transactions
                )

    def update(self, duration: float, save: bool = False, save_transactions: bool = False):
        """
        Update model for a single steps
        """
        # Delta
        if save is True:
            # Create current state
            previous_serialized = deepcopy(self.serialize())

        # Trade
        transactions = []
        for market_id in self.infrastructure.markets:  # Agents in market
            agents = self.society.agents_in(id=market_id)
            if len(agents) >= 1:
                transactions += self.trade(agents=agents, markets=[market_id])
        for home_id in self.infrastructure.homes:  # Agents in home
            agents = self.society.agents_in(id=home_id)
            if len(agents) >= 2:
                transactions += self.trade(agents=agents)
        #transactions
        for transaction in transactions:
            transaction.append(self.time)

        # Agents activity impact
        self.society.update(duration)

        # Climate impact
        self.infrastructure.update(duration)

        # Charge Markets (resource influx)
        markets = self.infrastructure.markets
        for id in markets:
            # Reset balance
            self.infrastructure.set_balance(id=id, value=0)
            # Reset resources
            for name in self.resource_names:
                enough_amount = self.infrastructure.get_enough_resource(id=id, name=name)
                self.infrastructure.set_resource(id=id, name=name, value=enough_amount)

        # General
        self.step += 1
        self.time += duration    

        # Delta
        if save is True:
            # Create new current state and compare it to the previous one
            current_serialized = self.serialize()
            delta = kd.create(
                old=previous_serialized,
                new=current_serialized
            )
            self.append_delta(delta)
            self.save_final()

        # Transactions
        if save_transactions is True:
            self.append_transactions(transactions)

        return transactions


class Print:

    def __str__(self):
        """
        Return print-friendly stats
        """
        stat = self.stat
        txt = ''
        for category in stat:
            for name in stat[category]:
                txt += f"# {name}: {str(stat[category][name])}" + "\n"
        txt = txt[:-1]
        return txt


from piperabm.economy.trade.trade_solver import MultiResourceTrade


def find_exchanges(players_initial, players_final):
    resources = ['food', 'water', 'energy']
    num = len(players_initial)
    sources = {}
    demands = {}
    for i in range(num):
        player_initial = players_initial[i]
        player_final = players_final[i]
        for resource in resources:
            player_initial_resource = player_initial['resources'][resource]
            player_final_resource = player_final['resources'][resource]
            diff = player_final_resource - player_initial_resource
            if resource not in sources:
                sources[resource] = []
            if resource not in demands:
                demands[resource] = []
            if diff < 0:
                sources[resource].append(-diff)
                demands[resource].append(0)
            else:
                sources[resource].append(0)
                demands[resource].append(diff)
    #print(sources)
    #print(demands)
    result = {}
    for resource in resources:
        resource_sources = sources[resource]
        resource_demands = demands[resource]
        transactions = resource_exchange(resource_sources, resource_demands)
        #result[resource] = transactions
        result[resource] = index_to_id(transactions, players_final)
    return result

def index_to_id(transactions, players):
    result = []
    #print(transactions[resource])
    for transaction in transactions:
        result.append(
            [
                players[transaction[0]]['id'], # from
                players[transaction[1]]['id'], # to
                transaction[2], # amount
            ]
        )
    return result

def resource_exchange(sources, demands):
    # Create a list of indices for sources and demands to keep track of the original positions
    source_indices = list(range(len(sources)))
    demand_indices = list(range(len(demands)))
    
    # We'll keep a log of transactions to see who gave to whom and how much
    transactions = []
    
    # Continue until we have no more sources or demands to fulfill
    while any(sources) and any(demands):
        # Get the index of the largest source and largest demand
        max_source_index = max(range(len(sources)), key=lambda x: sources[x])
        max_demand_index = max(range(len(demands)), key=lambda x: demands[x])
        
        # Find the maximum amount that can be transferred
        amount = min(sources[max_source_index], demands[max_demand_index])
        
        # Record the transaction
        transactions.append((source_indices[max_source_index], demand_indices[max_demand_index], amount))
        
        # Update the source and demand values
        sources[max_source_index] -= amount
        demands[max_demand_index] -= amount
        
        # If a source or demand is depleted, set it to 0 to avoid negative numbers
        if sources[max_source_index] < 1e-9:
            sources[max_source_index] = 0
        if demands[max_demand_index] < 1e-9:
            demands[max_demand_index] = 0
            
    return transactions


if __name__ == "__main__":

    from copy import deepcopy

    prices = {
        'food': 10,
        'water': 10,
        'energy': 10
    }
    player_1 = {
        'id': 1,
        'type': 'agent',
        'resources': {
            'food': 9,
            'water': 2,
            'energy': 3
        },
        'enough_resources': {
            'food': 10,
            'water': 10,
            'energy': 10
        },
        'balance': 100,
    }
    player_2 = {
        'id': 2,
        'type': 'agent',
        'resources': {
            'food': 10,
            'water': 3,
            'energy': 10
        },
        'enough_resources': {
            'food': 10,
            'water': 10,
            'energy': 10
        },
        'balance': 100,
    }
    player_3 = {
        'id': 3,
        'type': 'agent',
        'resources': {
            'food': 3,
            'water': 10,
            'energy': 7
        },
        'enough_resources': {
            'food': 10,
            'water': 10,
            'energy': 10
        },
        'balance': 10,
    }
    players = [player_1, player_2, player_3]

    players_initial = deepcopy(players)

    print("Solves multi-resources for multiple players at fixed price.")

    # Initial
    print("\n" + ">>> " + "Initial: ")
    for player in players:
        print(player['resources'], ', balance:', player['balance'])

    # Solve
    players = MultiResourceTrade.solve(players=players, prices=prices)

    # Final
    print("\n" + ">>> " + "Final: ")
    for player in players:
        print(player['resources'], ', balance:', player['balance'])

    # Echanges
    print("\n" + ">>> " + "Exchanges: ")
    exchanges = find_exchanges(players_initial=players_initial, players_final=players)
    print(exchanges)


import numpy as np

from piperabm.tools.vector import vector as vc


def point_to_point(point_1, point_2, vector=False, ndarray=False):
    result = None
    point_1 = np.array(point_1)
    point_2 = np.array(point_2)
    
    if vector is True:
        result = point_2 - point_1
        if ndarray is False:
            result = list(result)
    else:
        result = vc.magnitude(point_to_point(point_1, point_2, vector=True, ndarray=True))

    return result


if __name__ == '__main__':
    point_1 = [0, 0]
    point_2 = [3, 4]
    distance = point_to_point(point_1, point_2, vector=True)
    print(distance)